"""Test Redis cache in listener."""

import time

from resolwe.flow.managers.listener.redis_cache import (
    RedisLockStatus,
    cache_manager,
    redis_cache,
)
from resolwe.flow.models import Data, Process
from resolwe.test import TestCase


class RedisCache(TestCase):
    """Check that redis cache plugin is registered."""

    def setUp(self):
        """Clear the cache and prepare the objects."""
        result = super().setUp()
        redis_cache.clear()

        proc = Process.objects.create(
            type="data:test:process",
            slug="test-process",
            contributor=self.contributor,
        )
        self.data1 = Data.objects.create(
            name="Test data 1",
            contributor=self.contributor,
            process=proc,
            status=Data.STATUS_DONE,
        )
        self.data2 = Data.objects.create(
            name="Test data 2",
            contributor=self.contributor,
            process=proc,
            status=Data.STATUS_DONE,
        )
        return result

    def test_register_data(self):
        """Test that Data plugin is registered."""
        self.assertIn("flow.data", cache_manager._plugins)
        cache_manager.get_plugin(Data)

    def test_cache_single(self):
        """Test that object can be cached."""
        # Test that data is not cached.
        self.assertIsNone(cache_manager.get(Data, (self.data1.id,)))

        # Cache the data object and verify that it is cached.
        cache_manager.cache(self.data1)
        result = cache_manager.get(Data, (self.data1.id,))
        self.assertEqual(result["id"], self.data1.id)

    def test_cache_multi(self):
        """Test that multiple objects can be cached."""
        # Assert no cached data exists in Redis.
        cache = cache_manager.mget(Data, ((self.data1.id,), (self.data2.id,)))
        self.assertEqual(cache, [None, None])

        # Cache the data and verify that it is cached.
        cache_manager.mcache(Data.objects.filter(pk__in=[self.data1.pk, self.data2.pk]))
        cache = cache_manager.mget(Data, ((self.data1.id,), (self.data2.id,)))
        self.assertEqual(
            [entry["id"] for entry in cache], [self.data1.id, self.data2.id]
        )

    def test_clear(self):
        """Test that multiple objects can be cached."""
        # Cache the data and verify that it is cached.
        cache_manager.mcache(Data.objects.filter(pk__in=[self.data1.pk, self.data2.pk]))
        # Remove the first cache
        redis_cache.clear(Data, (self.data1.id,))
        cache = cache_manager.mget(Data, ((self.data1.id,), (self.data2.id,)))
        self.assertEqual(cache[0], None)
        self.assertEqual(cache[1]["id"], self.data2.id)

        redis_cache.clear(Data)
        cache = cache_manager.mget(Data, ((self.data1.id,), (self.data2.id,)))
        self.assertEqual(cache, [None, None])

        cache_manager.mcache(Data.objects.filter(pk__in=[self.data1.pk, self.data2.pk]))
        redis_cache.clear()
        self.assertEqual(cache, [None, None])

    def test_lock(self):
        """Test locking."""
        identifiers_list = [(self.data1.id,), (self.data2.id,)]
        # Without locking the wait should return immediately.
        start = time.time()
        result = cache_manager.wait(Data, identifiers_list, timeout=0.1)
        elapsed = time.time() - start
        self.assertLess(elapsed, 0.05)
        self.assertEqual(result, {None})

        # Lock the data1 and wait for it.
        cache_manager.lock(Data, [(self.data1.id,)])
        start = time.time()
        result = redis_cache.wait(
            Data, identifiers_list, timeout=0.1, refresh_interval=0.05
        )
        elapsed = time.time() - start
        self.assertTrue(0.1 < elapsed < 0.2)
        self.assertEqual(result, {None, RedisLockStatus.PROCESSING})

        # Verify the refresh interval is working.
        start = time.time()
        result = redis_cache.wait(
            Data, identifiers_list, timeout=0.1, refresh_interval=0.5
        )
        elapsed = time.time() - start
        self.assertTrue(0.5 < elapsed < 0.6)
        self.assertEqual(result, {None, RedisLockStatus.PROCESSING})

        # Unlocking data2 should not have any effect.
        cache_manager.unlock(Data, [(self.data2.id,)])
        start = time.time()
        result = redis_cache.wait(
            Data, identifiers_list, timeout=0.1, refresh_interval=0.05
        )
        elapsed = time.time() - start
        self.assertTrue(0.1 < elapsed < 0.2)
        self.assertEqual(result, {RedisLockStatus.OK, RedisLockStatus.PROCESSING})

        # Unlocking data1 should remove lock.
        cache_manager.unlock(Data, [(self.data1.id,)])
        start = time.time()
        result = cache_manager.wait(Data, identifiers_list, timeout=0.1)
        elapsed = time.time() - start
        self.assertLess(elapsed, 0.05)
        self.assertEqual(result, {RedisLockStatus.OK})

        # Try the lock unlocks itself after timeout.
        redis_cache.lock(Data, identifiers_list, valid_for=1)
        start = time.time()
        result = redis_cache.wait(
            Data, identifiers_list, timeout=2, refresh_interval=0.05
        )
        elapsed = time.time() - start
        self.assertEqual(result, {None})
        self.assertTrue(1 < elapsed < 1.1)



""".. Ignore pydocstyle D400.

==================
Flow Configuration
==================

"""
from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class FlowConfig(AppConfig):
    """Flow AppConfig."""

    name = "resolwe.flow"
    verbose_name = _("Resolwe Dataflow")

    def ready(self):
        """Application initialization."""
        # Register signals handlers
        from . import signals  # noqa: F401



"""Annotations viewset."""


from rest_framework import exceptions, mixins, permissions, request, viewsets

from resolwe.flow.filters import (
    AnnotationFieldFilter,
    AnnotationPresetFilter,
    AnnotationValueFilter,
)
from resolwe.flow.models import AnnotationPreset, Entity
from resolwe.flow.models.annotations import AnnotationField, AnnotationValue
from resolwe.flow.serializers.annotations import (
    AnnotationFieldSerializer,
    AnnotationPresetSerializer,
    AnnotationValueSerializer,
)
from resolwe.permissions.loader import get_permissions_class
from resolwe.permissions.mixins import ResolwePermissionsMixin
from resolwe.permissions.models import Permission

from .mixins import ResolweCreateModelMixin, ResolweUpdateModelMixin


class AnnotationPresetViewSet(
    ResolweCreateModelMixin,
    mixins.RetrieveModelMixin,
    ResolweUpdateModelMixin,
    mixins.ListModelMixin,
    ResolwePermissionsMixin,
    viewsets.GenericViewSet,
):
    """API view for :class:`DescriptorSchema` objects."""

    serializer_class = AnnotationPresetSerializer
    permission_classes = (get_permissions_class(),)
    filterset_class = AnnotationPresetFilter

    # No need to specify default ordering: it is specified on the model.
    ordering_fields = ("id", "name", "sort_order")
    queryset = AnnotationPreset.objects.all()

    def get_queryset(self):
        """Get the presets for the current user."""
        return super().get_queryset().filter_for_user(self.request.user)


class AnnotationFieldViewSet(
    mixins.RetrieveModelMixin,
    mixins.ListModelMixin,
    viewsets.GenericViewSet,
):
    """Annotation fields viewset."""

    permission_classes = (permissions.AllowAny,)
    serializer_class = AnnotationFieldSerializer
    filterset_class = AnnotationFieldFilter
    # No need to specify default ordering: it is specified on the model.
    ordering_fields = ("id", "name", "label", "group__sort_order", "sort_order")
    queryset = AnnotationField.objects.all()


class AnnotationValueViewSet(
    ResolweCreateModelMixin,
    mixins.RetrieveModelMixin,
    ResolweUpdateModelMixin,
    mixins.ListModelMixin,
    mixins.DestroyModelMixin,
    viewsets.GenericViewSet,
):
    """Annotation value viewset."""

    # Users can only see the annotation values on the entities they have permission to
    # access. The actual permissions are checked in the AnnotationValueFilter. The
    # filter assures at least one of the entity filters is applied and the permissions
    # filters are applied inside filter_permissions method defined in
    # AnnotationValueMetaclass.
    # This behaviour is tested in test_list_filter_values.
    permission_classes = (permissions.AllowAny,)
    serializer_class = AnnotationValueSerializer
    filterset_class = AnnotationValueFilter
    queryset = AnnotationValue.objects.all()

    def _has_permissions_on_entity(self, entity: Entity) -> bool:
        """Has the authenticated user EDIT permission on the associated entity."""
        return (
            Entity.objects.filter(pk=entity.pk)
            .filter_for_user(self.request.user, Permission.EDIT)
            .exists()
        )

    def _get_entity(self, request: request.Request) -> AnnotationValue:
        """Get annotation value from request.

        :raises ValidationError: if the annotation value is not valid.
        :raises NotFound: if the user is not authenticated.
        """
        if not request.user.is_authenticated:
            raise exceptions.NotFound

        serializer = self.get_serializer(data=request.data, partial=True)
        serializer.is_valid(raise_exception=True)
        return serializer.validated_data["entity"]

    def create(self, request, *args, **kwargs):
        """Create annotation value.

        Authenticated users with edit permissions on the entity can create annotations.
        """
        if self._has_permissions_on_entity(self._get_entity(request)):
            return super().create(request, *args, **kwargs)
        raise exceptions.NotFound()

    def update(self, request, *args, **kwargs):
        """Update annotation values.

        Authenticated users with edit permission on the entity can update annotations.
        """
        entity = AnnotationValue.objects.get(pk=kwargs["pk"]).entity
        if self._has_permissions_on_entity(entity):
            return super().update(request, *args, **kwargs)
        raise exceptions.NotFound()

    def destroy(self, request, *args, **kwargs):
        """Destroy the annotation value."""
        entity = AnnotationValue.objects.get(pk=kwargs["pk"]).entity
        if self._has_permissions_on_entity(entity):
            return super().destroy(request, *args, **kwargs)
        raise exceptions.PermissionDenied()



"""Base classes for plugin managenent."""

import abc
import logging
from typing import Generic, Type, TypeVar

from resolwe.utils import BraceMessage as __

logger = logging.getLogger(__name__)


PluginType = TypeVar("PluginType", bound="Plugin")


class PluginManager(Generic[PluginType], metaclass=abc.ABCMeta):
    """Base class for plugin manager."""

    def __init__(self):
        """Initialize the plugins dictionary."""
        self._plugins: dict[str, PluginType] = dict()

    def add_plugin(self, plugin_class: Type[PluginType]):
        """Add new plugin.

        :raises AssertionError: when plugin with the same identifier is already
            registered.
        """
        plugin_identifier = plugin_class.get_identifier()
        assert (
            plugin_identifier not in self._plugins
        ), f"Plugin '{plugin_identifier}' already registered."
        self._plugins[plugin_identifier] = plugin_class()
        logger.debug(__("Registering plugin '{}'.", plugin_identifier))

    def get_plugin(self, plugin_identifier: str) -> PluginType:
        """Get the plugin.

        :raises KeyError: when plugin is not registered.
        """
        return self._plugins[plugin_identifier]


class PluginMetaClass(abc.ABCMeta):
    """Metaclass for plugin classes."""

    def __new__(cls, name, bases, attrs):
        """Register the plugin when abstract not explicitily set to False."""
        plugin_class = super().__new__(cls, name, bases, attrs)
        if not attrs.get("abstract", False):
            plugin_class.plugin_manager.add_plugin(plugin_class)
        return plugin_class


class Plugin(metaclass=PluginMetaClass):
    """Basic plugin class.

    It defines method that registers the plugin with the manager at the creation.
    """

    # The name of the variable that represents a plugin manager to register with.
    plugin_manager: PluginManager
    abstract = True

    @classmethod
    def get_identifier(cls) -> str:
        """Return the plugin identifier."""
        return "{}.{}".format(cls.__module__, cls.__name__)



# pylint: disable=missing-docstring
from unittest.mock import patch

from django.contrib.auth import get_user_model

from resolwe.flow.executors import constants
from resolwe.flow.executors.socket_utils import Message, Response, ResponseStatus
from resolwe.flow.managers.listener.basic_commands_plugin import BasicCommands
from resolwe.flow.managers.listener.listener import Processor
from resolwe.flow.managers.protocol import ExecutorProtocol
from resolwe.flow.models import Data, DataDependency, Entity, Worker
from resolwe.flow.models.annotations import (
    AnnotationField,
    AnnotationGroup,
    AnnotationType,
    AnnotationValue,
)
from resolwe.flow.models.process import Process
from resolwe.permissions.models import Permission
from resolwe.storage.connectors.baseconnector import BaseStorageConnector
from resolwe.storage.connectors.s3connector import AwsS3Connector
from resolwe.storage.models import FileStorage, ReferencedPath, StorageLocation
from resolwe.test import TestCase


class ListenerTest(TestCase):
    fixtures = ["storage_data.yaml", "storage_processes.yaml", "storage_users.yaml"]

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.manager = Processor(None)
        cls.processor = BasicCommands()
        cls.file_storage = FileStorage.objects.get(id=1)
        cls.storage_location = StorageLocation.objects.create(
            file_storage=cls.file_storage, connector_name="GCS", status="OK"
        )
        cls.path = ReferencedPath.objects.create(
            path="test.me", md5="md5", crc32c="crc", awss3etag="aws"
        )
        cls.storage_location.files.add(cls.path)

    def test_handle_download_finished_missing_storage_location(self):
        obj = Message.command(ExecutorProtocol.DOWNLOAD_FINISHED, -2)
        with self.assertRaises(StorageLocation.DoesNotExist):
            self.processor.handle_download_finished(b"1", obj, self.manager)

    def test_handle_download_finished(self):
        storage_location = StorageLocation.objects.create(
            file_storage=self.file_storage, connector_name="local"
        )
        obj = Message.command(ExecutorProtocol.DOWNLOAD_FINISHED, storage_location.id)

        with patch(
            "resolwe.storage.models.FileStorage.default_storage_location",
            self.storage_location,
        ):
            response = self.processor.handle_download_finished(b"1", obj, self.manager)

        self.assertEqual(response.response_status, ResponseStatus.OK)
        storage_location.refresh_from_db()
        self.assertEqual(storage_location.status, StorageLocation.STATUS_DONE)
        self.assertEqual(storage_location.files.count(), 1)
        file = storage_location.files.get()
        self.assertEqual(file.path, "test.me")
        self.assertEqual(file.md5, "md5")
        self.assertEqual(file.crc32c, "crc")
        self.assertEqual(file.awss3etag, "aws")

    def test_handle_download_aborted_missing_storage_location(self):
        obj = Message.command(ExecutorProtocol.DOWNLOAD_ABORTED, -2)
        response = self.processor.handle_download_aborted(b"1", obj, self.manager)
        self.assertEqual(response.response_status, ResponseStatus.OK)

    def test_handle_download_aborted(self):
        storage_location = StorageLocation.objects.create(
            file_storage=self.file_storage,
            connector_name="local",
            status=StorageLocation.STATUS_UPLOADING,
        )
        obj = Message.command(ExecutorProtocol.DOWNLOAD_ABORTED, storage_location.id)
        self.processor.handle_download_aborted(b"1", obj, self.manager)

        storage_location.refresh_from_db()
        self.assertEqual(storage_location.status, StorageLocation.STATUS_PREPARING)

    def test_handle_download_started_no_location(self):
        obj = Message.command(
            ExecutorProtocol.DOWNLOAD_STARTED,
            {
                "storage_location_id": -2,
                "download_started_lock": True,
            },
        )
        with self.assertRaises(StorageLocation.DoesNotExist):
            self.processor.handle_download_started(b"1", obj, self.manager)

    def test_handle_download_started_ok_no_lock_preparing(self):
        storage_location = StorageLocation.objects.create(
            file_storage=self.file_storage, connector_name="local"
        )

        obj = Message.command(
            ExecutorProtocol.DOWNLOAD_STARTED,
            {
                "storage_location_id": storage_location.id,
                "download_started_lock": False,
            },
        )
        response = self.processor.handle_download_started(b"1", obj, self.manager)
        self.assertEqual(
            response, Response(ResponseStatus.OK.value, "download_started")
        )
        storage_location.refresh_from_db()
        self.assertEqual(storage_location.status, StorageLocation.STATUS_PREPARING)

    def test_handle_download_started_ok_no_lock_uploading(self):
        storage_location = StorageLocation.objects.create(
            file_storage=self.file_storage,
            connector_name="local",
            status=StorageLocation.STATUS_UPLOADING,
        )
        obj = Message.command(
            ExecutorProtocol.DOWNLOAD_STARTED,
            {
                "storage_location_id": storage_location.id,
                "download_started_lock": False,
            },
        )

        response = self.processor.handle_download_started(b"1", obj, self.manager)
        self.assertEqual(
            response, Response(ResponseStatus.OK.value, "download_in_progress")
        )
        storage_location.refresh_from_db()
        self.assertEqual(storage_location.status, StorageLocation.STATUS_UPLOADING)

    def test_handle_download_started_ok_no_lock_done(self):
        storage_location = StorageLocation.objects.create(
            file_storage=self.file_storage,
            connector_name="local",
            status=StorageLocation.STATUS_DONE,
        )
        obj = Message.command(
            ExecutorProtocol.DOWNLOAD_STARTED,
            {
                "storage_location_id": storage_location.id,
                "download_started_lock": False,
            },
        )
        response = self.processor.handle_download_started(b"1", obj, self.manager)
        self.assertEqual(
            response, Response(ResponseStatus.OK.value, "download_finished")
        )
        storage_location.refresh_from_db()
        self.assertEqual(storage_location.status, StorageLocation.STATUS_DONE)

    def test_handle_resolve_url(self):
        """Test resolve URL."""

        def resolve(url: str) -> str:
            """Resolve the given URL."""
            request = Message.command("resolve_url", url)
            return self.processor.handle_resolve_url(
                b"1", request, self.manager
            ).message_data

        def presigned_url_mock(key: str, expiration: int):
            """Patch for presigned URL method.

            Just return the given key prefixed by "presigned_".
            """
            return f"presigned_{key}"

        # When URL is not handled by the plugin it should not change.
        url = "http://test_me"
        self.assertEqual(resolve(url), url)
        url = "unsupported urls should not change."
        self.assertEqual(resolve(url), url)

        # Test URL resolving for S3 plugin.
        # When bucket is not known, the URL should not change.
        # When bucket is known, is must be resolved to the presigned url.
        s3connector = AwsS3Connector({"bucket": "test"}, "S3")
        s3connector.presigned_url = presigned_url_mock
        with patch(
            "resolwe.flow.managers.basic_commands_plugin.connectors",
            {"S3": s3connector},
        ):
            # Unknown bucket.
            url = "s3://unknown-bucket/resolve"
            self.assertEqual(resolve(url), url)
            # Known bucket.
            url = "s3://test/key"
            self.assertEqual(resolve(url), "presigned_key")

    def test_handle_download_started_ok_lock(self):
        storage_location = StorageLocation.objects.create(
            file_storage=self.file_storage, connector_name="local"
        )

        obj = Message.command(
            ExecutorProtocol.DOWNLOAD_STARTED,
            {
                "storage_location_id": storage_location.id,
                "download_started_lock": True,
            },
        )
        response = self.processor.handle_download_started(b"1", obj, self.manager)
        self.assertEqual(
            response, Response(ResponseStatus.OK.value, "download_started")
        )
        storage_location.refresh_from_db()
        self.assertEqual(storage_location.status, StorageLocation.STATUS_UPLOADING)

    def test_handle_get_files_to_download_missing_storage_location(self):
        obj = Message.command(ExecutorProtocol.GET_FILES_TO_DOWNLOAD, -2)
        response = self.processor.handle_get_files_to_download(b"1", obj, self.manager)
        self.assertEqual(response, Response(ResponseStatus.OK.value, []))

    def test_handle_get_files_to_download(self):
        obj = Message.command(
            ExecutorProtocol.GET_FILES_TO_DOWNLOAD, self.storage_location.id
        )
        response = self.processor.handle_get_files_to_download(b"1", obj, self.manager)
        expected = Response(
            ResponseStatus.OK.value,
            [
                {
                    "id": self.path.id,
                    "path": "test.me",
                    "size": -1,
                    "md5": "md5",
                    "crc32c": "crc",
                    "awss3etag": "aws",
                    "chunk_size": BaseStorageConnector.CHUNK_SIZE,
                }
            ],
        )
        self.assertEqual(response, expected)

    def test_handle_missing_data_locations_missing_data(self):
        obj = Message.command(ExecutorProtocol.MISSING_DATA_LOCATIONS, "")
        response = self.processor.handle_missing_data_locations(b"1", obj, self.manager)
        self.assertEqual(response, Response(ResponseStatus.OK.value, {}))

    def test_handle_missing_data_locations_missing_storage_location(self):
        obj = Message.command(ExecutorProtocol.MISSING_DATA_LOCATIONS, "")
        parent = Data.objects.get(id=2)
        child = Data.objects.get(id=1)
        DataDependency.objects.create(
            parent=parent, child=child, kind=DataDependency.KIND_IO
        )
        response = self.processor.handle_missing_data_locations(b"1", obj, self.manager)
        expected = Response(ResponseStatus.ERROR.value, "No storage location exists")
        self.assertEqual(response, expected)
        self.assertEqual(StorageLocation.all_objects.count(), 1)

    def test_handle_missing_data_locations_none(self):
        obj = Message.command(ExecutorProtocol.MISSING_DATA_LOCATIONS, "")
        parent = Data.objects.get(id=2)
        child = Data.objects.get(id=1)
        DataDependency.objects.create(
            parent=parent, child=child, kind=DataDependency.KIND_IO
        )
        StorageLocation.objects.create(
            file_storage=parent.location,
            connector_name="local",
            status=StorageLocation.STATUS_DONE,
            url="url",
        )
        response = self.processor.handle_missing_data_locations(b"1", obj, self.manager)
        expected = Response(ResponseStatus.OK.value, {})
        self.assertEqual(response, expected)
        self.assertEqual(StorageLocation.all_objects.count(), 2)

    def test_handle_missing_data_locations(self):
        obj = Message.command(ExecutorProtocol.MISSING_DATA_LOCATIONS, "")
        parent = Data.objects.get(id=2)
        child = Data.objects.get(id=1)
        DataDependency.objects.create(
            parent=parent, child=child, kind=DataDependency.KIND_IO
        )
        storage_location = StorageLocation.objects.create(
            file_storage=parent.location,
            connector_name="not_local",
            status=StorageLocation.STATUS_DONE,
            url="url",
        )
        response = self.processor.handle_missing_data_locations(b"1", obj, self.manager)
        self.assertEqual(StorageLocation.all_objects.count(), 3)
        created = StorageLocation.all_objects.last()
        expected = Response(
            ResponseStatus.OK.value,
            {
                "url": {
                    "data_id": parent.id,
                    "from_connector": "not_local",
                    "from_storage_location_id": storage_location.id,
                    "to_storage_location_id": created.id,
                    "to_connector": "local",
                }
            },
        )
        self.assertEqual(response, expected)

    def test_handle_resolve_data_path(self):
        """Test data path resolving."""
        file_storage = FileStorage.objects.create()
        process = Process.objects.create(contributor=self.contributor)
        data = Data.objects.create(
            name="Test min", process=process, contributor=self.contributor
        )
        data.status = Data.STATUS_PROCESSING
        data.location_id = file_storage.id
        data.save()
        Worker.objects.get_or_create(data_id=data.pk, status=Worker.STATUS_PROCESSING)

        peer_identity = str(data.pk).encode()
        message = Message.command("resolve_data_path", data.pk)
        response = self.manager.process_command(peer_identity, message)
        self.assertEqual(response.message_data, str(constants.INPUTS_VOLUME))
        connector_name = "local"
        self.storage_location = StorageLocation.objects.create(
            file_storage=file_storage, connector_name="local", status="OK"
        )
        # Generate new command uuid or the processing will be skipped.
        message = Message.command("resolve_data_path", data.pk)
        response = self.manager.process_command(peer_identity, message)
        self.assertEqual(response.message_data, f"/data_{connector_name}")

    def test_get_entity_annotations(self):
        """Test entity annotations retrieval."""
        user_model = get_user_model()
        user_model.objects.create_user(
            username="public",
            email="public@test.com",
            password="public",
            first_name="James",
            last_name="Smith",
        )
        entity = Entity.objects.create(name="Entity", contributor=self.contributor)

        group = AnnotationGroup.objects.create(name="group", sort_order=1)
        field = AnnotationField.objects.create(
            name="field",
            type=AnnotationType.STRING.value,
            group=group,
            required=False,
            sort_order=1,
        )
        field2 = AnnotationField.objects.create(
            name="another field",
            type=AnnotationType.STRING.value,
            group=group,
            required=False,
            sort_order=2,
        )
        AnnotationValue.objects.create(entity=entity, field=field, value="value")
        AnnotationValue.objects.create(entity=entity, field=field2, value="value2")

        process = Process.objects.create(contributor=self.contributor)
        data = Data.objects.create(
            name="Test min", process=process, contributor=self.contributor
        )
        data.status = Data.STATUS_PROCESSING
        data.save()
        peer_identity = str(data.pk).encode()
        Worker.objects.update_or_create(
            data_id=data.pk, status=Worker.STATUS_PROCESSING
        )

        # Request without permisions should fail.
        message = Message.command(
            "get_entity_annotations",
            [entity.pk, ["group.field", "group.another field", "nonexisting.field"]],
        )
        expected = Response(
            ResponseStatus.ERROR.value,
            "Error in command handler 'handle_get_entity_annotations'.",
        )
        response = self.manager.process_command(peer_identity, message)
        self.assertEqual(response, expected)

        data = Data.objects.create(
            name="Test min", process=process, contributor=self.contributor
        )

        # Request with permissions should succeed and return only existing annotations.
        data.status = Data.STATUS_PROCESSING
        data.save()
        peer_identity = str(data.pk).encode()
        Worker.objects.update_or_create(
            data_id=data.pk, status=Worker.STATUS_PROCESSING
        )
        entity.set_permission(Permission.VIEW, self.contributor)
        response = self.manager.process_command(peer_identity, message)
        expected = Response(
            ResponseStatus.OK.value,
            {"group.field": "value", "group.another field": "value2"},
        )
        self.assertEqual(response, expected)

        # Request without annotation names should return all annotations.
        message = Message.command("get_entity_annotations", [entity.pk, None])
        response = self.manager.process_command(peer_identity, message)
        expected = Response(
            ResponseStatus.OK.value,
            {"group.field": "value", "group.another field": "value2"},
        )
        self.assertEqual(response, expected)

    def test_set_entity_annotations(self):
        """Test entity annotations creation / update."""

        user_model = get_user_model()
        user_model.objects.create_user(
            username="public",
            email="public@test.com",
            password="public",
            first_name="James",
            last_name="Smith",
        )
        entity = Entity.objects.create(name="Entity", contributor=self.contributor)

        group = AnnotationGroup.objects.create(name="group", sort_order=1)
        AnnotationField.objects.create(
            name="field",
            type=AnnotationType.STRING.value,
            group=group,
            required=False,
            sort_order=1,
        )
        AnnotationField.objects.create(
            name="another field",
            type=AnnotationType.STRING.value,
            group=group,
            required=False,
            sort_order=2,
        )

        process = Process.objects.create(contributor=self.contributor)
        data = Data.objects.create(
            name="Test min", process=process, contributor=self.contributor
        )
        data.status = Data.STATUS_PROCESSING
        data.save()
        peer_identity = str(data.pk).encode()
        Worker.objects.update_or_create(
            data_id=data.pk, status=Worker.STATUS_PROCESSING
        )

        # Request without permisions should fail.
        message = Message.command(
            "set_entity_annotations", [entity.pk, {"group.field": "value"}, True]
        )
        expected = Response(
            ResponseStatus.ERROR.value,
            "Error in command handler 'handle_set_entity_annotations'.",
        )
        response = self.manager.process_command(peer_identity, message)
        self.assertEqual(response, expected)

        # Request with permissions should succeed.
        data = Data.objects.create(
            name="Test min", process=process, contributor=self.contributor
        )
        data.status = Data.STATUS_PROCESSING
        data.save()
        peer_identity = str(data.pk).encode()
        Worker.objects.update_or_create(
            data_id=data.pk, status=Worker.STATUS_PROCESSING
        )
        entity.set_permission(Permission.EDIT, self.contributor)
        self.assertEqual(entity.annotations.count(), 0)

        response = self.manager.process_command(peer_identity, message)
        expected = Response(ResponseStatus.OK.value, "OK")
        self.assertEqual(entity.annotations.count(), 1)
        self.assertAnnotation(entity, "group.field", "value")

        # Annotations should be updated or created
        message = Message.command(
            "set_entity_annotations",
            [
                entity.pk,
                {"group.field": "updated value", "group.another field": "value2"},
                True,
            ],
        )
        response = self.manager.process_command(peer_identity, message)
        expected = Response(ResponseStatus.OK.value, "OK")
        self.assertEqual(entity.annotations.count(), 2)
        self.assertAnnotation(entity, "group.field", "updated value")
        self.assertAnnotation(entity, "group.another field", "value2")



""".. Ignore pydocstyle D400.

==================
Permissions Filter
==================

"""
from rest_framework.filters import BaseFilterBackend

from resolwe.permissions.utils import model_has_permissions


class ResolwePermissionsFilter(BaseFilterBackend):
    """Permissions filter."""

    def filter_queryset(self, request, queryset, view):
        """Filter permissions queryset.

        When model has no permission group defined return the entire queryset.
        """
        if model_has_permissions(queryset.model):
            return queryset.filter_for_user(request.user)
        else:
            return queryset



""".. Ignore pydocstyle D400.

====================
Kubernetes Connector
====================

"""
import hashlib
import json
import logging
import os
import re
import time
from contextlib import suppress
from enum import Enum
from importlib.util import find_spec
from pathlib import Path
from typing import Any, Dict, Iterable, Tuple

import kubernetes

from django.conf import settings
from django.db.models import Sum
from django.db.models.functions import Coalesce

from resolwe.flow.executors import constants
from resolwe.flow.models import Data, DataDependency, Process
from resolwe.flow.utils.decorators import retry
from resolwe.storage import settings as storage_settings
from resolwe.storage.connectors import connectors
from resolwe.storage.connectors.baseconnector import BaseStorageConnector
from resolwe.utils import BraceMessage as __

from .base import BaseConnector

# TODO: is this really needed?
# Limits of containers' access to memory. We set the limit to ensure
# processes are stable and do not get killed by OOM signal.
KUBERNETES_MEMORY_HARD_LIMIT_BUFFER = 2000

# Timeout (in seconds) to wait for response from kubernetes API.
KUBERNETES_TIMEOUT = 30

logger = logging.getLogger(__name__)


def get_mountable_connectors() -> Iterable[Tuple[str, BaseStorageConnector]]:
    """Iterate through all the storages and find mountable connectors.

    :returns: list of tuples (storage_name, connector).
    """
    return [
        (storage_name, connector)
        for storage_name in storage_settings.FLOW_STORAGE
        for connector in connectors.for_storage(storage_name)
        if connector.mountable
    ]


def get_upload_dir() -> str:
    """Get the upload path.

    : returns: the path of the first mountable connector for storage
        'upload'.

    :raises RuntimeError: if no applicable connector is found.
    """
    for connector in connectors.for_storage("upload"):
        if connector.mountable:
            return f"/upload_{connector.name}"
    raise RuntimeError("No mountable upload connector is defined.")


def unique_volume_name(base_name: str, data_id: int) -> str:
    """Get unique persistent volume claim name."""
    return f"{base_name}-{data_id}"


def sanitize_kubernetes_label(label: str, trim_end: bool = True) -> str:
    """Make sure kubernetes label complies with the rules.

    See the URL bellow for details.

    https://kubernetes.io/docs/concepts/overview/working-with-objects/labels/
    """
    max_length = 63
    sanitized_label = re.sub("[^0-9a-zA-Z\-]+", "-", label).strip("-_.")
    if len(sanitized_label) > max_length:
        logger.warning(__("Label '%s' is too long and was truncated.", label))
        if trim_end:
            sanitized_label = sanitized_label[-max_length:].strip("-_.")
        else:
            sanitized_label = sanitized_label[:max_length].strip("-_.")
    return sanitized_label


class ConfigLocation(Enum):
    """The enum specifying where to read the configuration from."""

    INCLUSTER = "incluster"
    KUBECTL = "kubectl"


class Connector(BaseConnector):
    """Kubernetes-based connector for job execution."""

    def __init__(self):
        """Initialize."""
        self._config_location = ConfigLocation(
            settings.KUBERNETES_DISPATCHER_CONFIG_LOCATION
        )
        self._initialize_variables()

    def _initialize_variables(self):
        """Init variables.

        This has to be done for every run since settings values may be overriden for tests.
        """
        self.kubernetes_namespace = getattr(settings, "KUBERNETES_SETTINGS", {}).get(
            "namespace", "default"
        )
        self.tools_path_prefix = Path("/usr/local/bin/resolwe")

    def _prepare_environment(
        self,
        data: Data,
        listener_connection: Tuple[str, str, str],
        tools_configmaps: Dict[str, str],
    ) -> list:
        """Prepare environmental variables."""
        host, port, protocol = listener_connection
        processing_uid, processing_gid = self._get_processing_uid_gid()
        environment = {
            "LISTENER_SERVICE_HOST": host,
            "LISTENER_SERVICE_PORT": port,
            "LISTENER_PROTOCOL": protocol,
            "DATA_ID": data.id,
            "FLOW_MANAGER_KEEP_DATA": getattr(
                settings, "FLOW_MANAGER_KEEP_DATA", False
            ),
            "RUNNING_IN_CONTAINER": 1,
            "RUNNING_IN_KUBERNETES": 1,
            "GENIALIS_UID": processing_uid,
            "GENIALIS_GID": processing_gid,
            "DESCRIPTOR_CHUNK_SIZE": 100,
            "MOUNTED_CONNECTORS": ",".join(
                connector.name
                for connector in connectors.values()
                if connector.mountable
            ),
            "TOOLS_PATHS": ":".join(
                f"{self.tools_path_prefix / tools_name}"
                for tools_name in tools_configmaps
            ),
        }
        with suppress(RuntimeError):
            environment["UPLOAD_DIR"] = get_upload_dir()

        return [
            {"name": name, "value": str(value)} for name, value in environment.items()
        ]

    def _create_configmap_if_needed(self, name: str, content: Dict, core_api: Any):
        """Create configmap if necessary."""
        try:
            core_api.read_namespaced_config_map(
                name=name, namespace=self.kubernetes_namespace
            )
        except kubernetes.client.rest.ApiException:
            # The configmap is not found, create one.
            configmap_description = {
                "apiVersion": "v1",
                "kind": "ConfigMap",
                "metadata": {"name": name},
                "data": content,
            }
            # The configmap might have already been created in the meantime.
            with suppress(kubernetes.client.rest.ApiException):
                core_api.create_namespaced_config_map(
                    body=configmap_description,
                    namespace=self.kubernetes_namespace,
                    _request_timeout=KUBERNETES_TIMEOUT,
                )

    def _get_tools_configmaps(self, core_api) -> Dict[str, str]:
        """Get and return configmaps for tools."""
        description_configmap_name = (
            getattr(settings, "KUBERNETES_TOOLS_CONFIGMAPS", None) or "tools-configmaps"
        )
        configmap = core_api.read_namespaced_config_map(
            name=description_configmap_name, namespace=self.kubernetes_namespace
        )
        return configmap.data

    def _get_files_configmap_name(self, location_subpath: Path, core_api: Any):
        """Get or create configmap for files.

        We have to map group, passwd and startup files inside containers.
        This is done using configmap which is mostly static. This command
        returns description for this configmap.
        """
        processing_uid, processing_gid = self._get_processing_uid_gid()
        passwd_content = "root:x:0:0:root:/root:/bin/bash\n"
        passwd_content += f"user:x:{processing_uid}:{processing_gid}:user:{os.fspath(constants.PROCESSING_VOLUME)}:/bin/bash\n"
        group_content = "root:x:0:\n"
        group_content += f"user:x:{processing_gid}:user\n"

        data = dict()
        modules = {
            "communicator": "resolwe.process.communicator",
            "bootstrap-python-runtime": "resolwe.process.bootstrap_python_runtime",
            "socket-utils": "resolwe.flow.executors.socket_utils",
            "constants": "resolwe.flow.executors.constants",
            "startup-script": "resolwe.flow.executors.startup_processing_container",
        }
        for module, full_module_name in modules.items():
            spec = find_spec(full_module_name)
            assert (
                spec is not None and spec.origin is not None
            ), f"Unable to determine module {full_module_name} source code."
            data[module] = Path(spec.origin).read_text()

        data.update(
            {
                "passwd": passwd_content,
                "group": group_content,
            }
        )

        data_md5 = hashlib.md5(json.dumps(data, sort_keys=True).encode()).hexdigest()
        configmap_name = f"configmap-files-{data_md5}"

        self._create_configmap_if_needed(configmap_name, data, core_api)
        return configmap_name

    def _volumes(
        self,
        data_id: int,
        location_subpath: Path,
        core_api: Any,
        tools_configmaps: Dict[str, str],
    ) -> list:
        """Prepare all volumes."""

        def volume_from_config(volume_name: str, volume_config: dict):
            """Get configuration for kubernetes for given volume."""
            claim_name = volume_config["config"]["name"]
            if self._should_create_pvc(volume_config):
                claim_name = unique_volume_name(claim_name, data_id)
            if volume_config["type"] == "persistent_volume":
                return {
                    "name": volume_name,
                    "persistentVolumeClaim": {"claimName": claim_name},
                }
            elif volume_config["type"] == "host_path":
                return {
                    "name": volume_name,
                    "hostPath": {"path": os.fspath(volume_config["config"]["path"])},
                }
            elif volume_config["type"] == "temporary_directory":
                return {"name": volume_name, "emptyDir": {}}
            else:
                raise RuntimeError(
                    f"Unsupported volume configuration: {volume_config}."
                )

        files_configmap_name = self._get_files_configmap_name(
            location_subpath, core_api
        )

        volumes = [
            {
                "name": "files-volume",
                "configMap": {"name": files_configmap_name},
            }
        ]

        for tools_name, configmap_name in tools_configmaps.items():
            volumes.append(
                {
                    "name": f"tools-{tools_name}",
                    "configMap": {"name": configmap_name, "defaultMode": 0o755},
                }
            )

        volumes += [
            volume_from_config(volume_name, volume_config)
            for volume_name, volume_config in storage_settings.FLOW_VOLUMES.items()
            if volume_config["config"]["name"] != "tools"
        ]

        for storage_name, connector in get_mountable_connectors():
            claim_name = connector.config.get("persistent_volume_claim", None)
            volume_data: Dict[str, Any] = {
                "name": connector.name,
            }
            if claim_name:
                volume_data["persistentVolumeClaim"] = {"claimName": claim_name}
            else:
                volume_data["hostPath"] = (
                    {"path": os.fspath(connector.config["path"])},
                )
            volumes.append(volume_data)

        return volumes

    def _init_container_mountpoints(self):
        """Prepare mountpoints for init container.

        Processing and input volume (if defined) and all mountable connectors
        are mounted inside container.
        """
        mount_points = [
            {
                "name": constants.PROCESSING_VOLUME_NAME,
                "mountPath": os.fspath(constants.PROCESSING_VOLUME),
                "readOnly": False,
            },
            {
                "name": constants.SECRETS_VOLUME_NAME,
                "mountPath": os.fspath(constants.SECRETS_VOLUME),
                "readOnly": False,
            },
        ]
        if constants.INPUTS_VOLUME_NAME in storage_settings.FLOW_VOLUMES:
            mount_points.append(
                {
                    "name": constants.INPUTS_VOLUME_NAME,
                    "mountPath": os.fspath(constants.INPUTS_VOLUME),
                    "readOnly": False,
                }
            )
        mount_points += [
            {
                "name": connector.name,
                "mountPath": f"/{storage_name}_{connector.name}",
                "readOnly": False,
            }
            for storage_name, connector in get_mountable_connectors()
        ]
        return mount_points

    def _communicator_mountpoints(self, location_subpath: Path) -> list:
        """Mountpoints for communicator container.

        Socket directory and mountable connectors are mounted inside.
        """
        mount_points = [
            {
                "name": constants.SOCKETS_VOLUME_NAME,
                "mountPath": os.fspath(constants.SOCKETS_VOLUME),
                "readOnly": False,
            },
            {
                "name": constants.SECRETS_VOLUME_NAME,
                "mountPath": os.fspath(constants.SECRETS_VOLUME),
                "readOnly": False,
            },
        ]
        mount_points += [
            {
                "name": connector.name,
                "mountPath": f"/{storage_name}_{connector.name}",
                "readOnly": False,
            }
            for storage_name, connector in get_mountable_connectors()
        ]
        return mount_points

    def _processing_mountpoints(
        self, location_subpath: Path, tools_configmaps: Dict[str, str]
    ):
        """Mountpoints for processing container.

        Processing and input volume (if defined) and all mountable connectors
        are mounted inside container. All except processing volume are mounted
        read-only.
        """
        mount_points = [
            {
                "name": constants.PROCESSING_VOLUME_NAME,
                "mountPath": os.fspath(constants.PROCESSING_VOLUME),
                "subPath": os.fspath(location_subpath),
                "readOnly": False,
            },
        ]
        if constants.INPUTS_VOLUME_NAME in storage_settings.FLOW_VOLUMES:
            mount_points.append(
                {
                    "name": constants.INPUTS_VOLUME_NAME,
                    "mountPath": os.fspath(constants.INPUTS_VOLUME),
                    "readOnly": False,
                }
            )
        mount_points += [
            {
                "name": connector.name,
                "mountPath": f"/{storage_name}_{connector.name}",
                "readOnly": storage_name != "upload",
            }
            for storage_name, connector in get_mountable_connectors()
        ]

        mount_points += [
            {
                "name": "files-volume",
                "mountPath": "/etc/passwd",
                "subPath": "passwd",
            },
            {
                "name": "files-volume",
                "mountPath": "/etc/group",
                "subPath": "group",
            },
            {
                "name": "files-volume",
                "mountPath": "/socket_utils.py",
                "subPath": "socket-utils",
            },
            {
                "name": "files-volume",
                "mountPath": "/processing.py",
                "subPath": "startup-script",
            },
            {
                "name": "files-volume",
                "mountPath": "/constants.py",
                "subPath": "constants",
            },
            {
                "name": "files-volume",
                "mountPath": f"/{constants.BOOTSTRAP_PYTHON_RUNTIME}",
                "subPath": "bootstrap-python-runtime",
            },
            {
                "name": "files-volume",
                "mountPath": "/communicator.py",
                "subPath": "communicator",
            },
            {
                "name": constants.SOCKETS_VOLUME_NAME,
                "mountPath": os.fspath(constants.SOCKETS_VOLUME),
            },
            {
                "name": constants.SECRETS_VOLUME_NAME,
                "mountPath": os.fspath(constants.SECRETS_VOLUME),
                "readOnly": True,
            },
        ]
        for tool_name in tools_configmaps:
            mount_points.append(
                {
                    "name": f"tools-{tool_name}",
                    "mountPath": f"{self.tools_path_prefix / tool_name}",
                }
            )
        return mount_points

    def _persistent_volume_claim(
        self, claim_name: str, size: int, volume_config: Dict
    ) -> Dict[str, Any]:
        """Prepare claim for persistent volume."""
        return {
            "apiVersion": "v1",
            "kind": "PersistentVolumeClaim",
            "metadata": {"name": claim_name},
            "spec": {
                "accessModes": ["ReadWriteOnce"],
                "storageClassName": volume_config.get("storageClassName", "gp2"),
                "resources": {
                    "requests": {
                        "storage": size,
                    }
                },
            },
        }

    def _get_processing_uid_gid(self) -> Tuple[int, int]:
        """Get the UID and GID of the processing container."""
        return getattr(settings, "FLOW_PROCESSING_USER", os.getuid()), getattr(
            settings, "FLOW_PROCESSING_GROUP", os.getgid()
        )

    def _data_inputs_size(self, data: Data, safety_buffer: int = 2**30) -> int:
        """Get the size of data inputs.

        Also add 10% of the volume size + 2GB as safety buffer. When having a
        large filesystem with lots of files the filesystem overhead can be
        significant so a fixed safety buffer alone is not sufficient.

        :returns: the size of the input data in bytes.
        """
        inputs_size = (
            Data.objects.filter(
                children_dependency__child=data,
                children_dependency__kind=DataDependency.KIND_IO,
            )
            .aggregate(total_size=Coalesce(Sum("size"), 0))
            .get("total_size")
        )
        assert isinstance(inputs_size, int)
        return int(1.1 * (inputs_size + safety_buffer))

    def _should_create_pvc(self, volume_config: dict) -> bool:
        """Return True if pvc for the given volume should be created."""
        if volume_config["type"] != "persistent_volume":
            return False

        return volume_config["config"].get("create_pvc", False)

    def _image_mapper(self, image_name: str, mapper: Dict[str, str]) -> str:
        """Transform the image name if necessary.

        When image_name starts with one of the keys defined in the mapper the
        matching part is replaced with the corresponding value.
        """
        for key in mapper:
            if image_name.startswith(key):
                return f"{mapper[key]}{image_name[len(key):]}"
        return image_name

    @retry(
        logger=logger,
        max_retries=5,
        retry_exceptions=(kubernetes.config.config_exception.ConfigException,),
        cleanup_callback=lambda: kubernetes.config.kube_config._cleanup_temp_files(),
    )
    def _load_kubernetes_config(self):
        """Load the kubernetes configuration.

        :raises kubernetes.config.config_exception.ConfigException: when the
            configuration could not be read.
        """
        (
            kubernetes.config.load_incluster_config()
            if self._config_location == ConfigLocation.INCLUSTER
            else kubernetes.config.load_kube_config()
        )

    def optimize_job_scheduling(self, data: Data, job_description: dict):
        """Optimize the scheduling by modifying the job description.

        Moodify the job description to be a better fit on the scheduler. Currently this
        is used to schedule based on the process scheduling class.
        """
        scheduling_class_name = dict(Process.SCHEDULING_CLASS_CHOICES)[
            data.process.scheduling_class
        ].lower()
        if kubernetes_affinity := getattr(settings, "FLOW_KUBERNETES_AFFINITY", None):
            scheduling_class_affinity = kubernetes_affinity.get("scheduling_class", {})
            affinity_settings = scheduling_class_affinity.get(scheduling_class_name, {})
            if match_expressions := [
                {"key": key, "operator": "In", "values": values}
                for key, values in affinity_settings.items()
            ]:
                job_description["spec"]["template"]["spec"]["affinity"] = {
                    "nodeAffinity": {
                        "requiredDuringSchedulingIgnoredDuringExecution": {
                            "nodeSelectorTerms": [
                                {"matchExpressions": match_expressions}
                            ]
                        }
                    }
                }

    def _get_overcommit_factors(self, data: Data) -> dict:
        """Get the overcommit settings for CPU and memory.

        The returned dict is of the form:
        { 'cpu': 0.8, 'memory': 0.8 }

        The numbers indicite that we will only request 80% of CPU and RAM requested in
        the process.
        """
        default = {"cpu": 0.8, "memory": 0.8}
        job_type = dict(Process.SCHEDULING_CLASS_CHOICES)[data.process.scheduling_class]
        return default | settings.FLOW_KUBERNETES_OVERCOMMIT.get(job_type, {})

    def start(self, data: Data, listener_connection: Tuple[str, str, str]):
        """Start process execution.

        Construct kubernetes job description and pass it to the kubernetes.
        """
        try:
            self._load_kubernetes_config()
        except kubernetes.config.config_exception.ConfigException as exception:
            logger.exception("Could not load the kubernetes configuration.")
            raise exception

        batch_api = kubernetes.client.BatchV1Api()
        core_api = kubernetes.client.CoreV1Api()

        tools_configmaps = self._get_tools_configmaps(core_api)
        container_environment = self._prepare_environment(
            data, listener_connection, tools_configmaps
        )
        location_subpath = Path(data.location.subpath)

        container_name_prefix = (
            getattr(settings, "FLOW_EXECUTOR", {})
            .get("CONTAINER_NAME_PREFIX", "resolwe")
            .replace("_", "-")
            .lower()
        )
        container_name = self._generate_container_name(container_name_prefix, data.pk)

        # Set resource limits.
        requests = dict()
        limits = data.get_resource_limits()
        overcommit_factors = self._get_overcommit_factors(data)

        requests["cpu"] = limits["cores"] * overcommit_factors["cpu"]
        limits["cpu"] = limits.pop("cores") + 1

        # The memory in the database is stored in megabytes but the kubertenes
        # requires memory in bytes.
        # We request less memory than stored in the database and set limit at 10% more
        # plus KUBERNETES_MEMORY_HARD_LIMIT_BUFFER. The processes usually require 16GB,
        # 32GB... and since the node usualy has 64GB of memory and some of it is
        # consumed by the system processes only one process process that requires 32GB
        # can run on a node instead of 2.

        requests["memory"] = limits["memory"] * overcommit_factors["memory"]
        limits["memory"] = 1.1 * limits["memory"] + KUBERNETES_MEMORY_HARD_LIMIT_BUFFER
        limits["memory"] *= 2**20  # 2 ** 20 = mebibyte
        requests["memory"] *= 2**20

        # Get the limits and requests for the communicator container.
        communicator_limits = getattr(
            settings,
            "FLOW_KUBERNETES_COMMUNICATOR_LIMITS",
            {"memory": "256M", "cpu": 0.1},
        )
        communicator_requests = getattr(
            settings,
            "FLOW_KUBERNETES_COMMUNICATOR_REQUESTS",
            {"memory": "256M", "cpu": 0.1},
        )

        resources = data.process.requirements.get("resources", {})
        network = "bridge"
        use_host_network = False
        if "network" in resources:
            # Configure Docker network mode for the container (if specified).
            # By default, current Docker versions use the 'bridge' mode which
            # creates a network stack on the default Docker bridge.
            network = getattr(settings, "FLOW_EXECUTOR", {}).get("NETWORK", "")
            use_host_network = network == "host"

        # Generate and set seccomp policy to limit syscalls.
        processing_uid, processing_gid = self._get_processing_uid_gid()
        security_context = {
            "runAsUser": processing_uid,
            "runAsGroup": processing_gid,
            "allowPrivilegeEscalation": False,
            "privileged": False,
            "capabilities": {"drop": ["ALL"]},
        }

        annotations = dict()

        # Do not evict job from node.
        annotations["cluster-autoscaler.kubernetes.io/safe-to-evict"] = "false"

        # The path is a relative path in the kubelet root
        # directory:
        # <seccomp_root>/<path>, where <seccomp_root> is defined via the
        # --seccomp-profile-root flag on the Kubelet. If the
        # --seccomp-profile-root flag is not defined, the default path will
        # be used, which is <root-dir>/seccomp where <root-dir> is
        # specified by the --root-dir flag.
        # https://kubernetes.io/docs/concepts/policy/pod-security-policy/
        #
        # The file is transfered to kubelets with daemonset ? Currently I
        # mount my /tmp directory to the /seccomp directory in minikube.
        annotations["seccomp.security.alpha.kubernetes.io/pod"] = "runtime/default"

        mapper = getattr(settings, "FLOW_CONTAINER_IMAGE_MAP", {})
        communicator_image = getattr(
            settings,
            "FLOW_DOCKER_COMMUNICATOR_IMAGE",
            "public.ecr.aws/s4q6j6e8/resolwe/com:latest",
        )
        communicator_image = self._image_mapper(communicator_image, mapper)

        requirements = data.process.requirements.get("executor", {}).get("docker", {})
        processing_container_image = str(
            requirements.get(
                "image",
                getattr(
                    settings,
                    "FLOW_DOCKER_DEFAULT_PROCESSING_CONTAINER_IMAGE",
                    "public.ecr.aws/s4q6j6e8/resolwe/base:ubuntu-20.04",
                ),
            ),
        )
        processing_container_image = self._image_mapper(
            processing_container_image, mapper
        )

        job_type = dict(Process.SCHEDULING_CLASS_CHOICES)[data.process.scheduling_class]
        job_description = {
            "apiVersion": "batch/v1",
            "kind": "Job",
            "metadata": {"name": sanitize_kubernetes_label(container_name)},
            "spec": {
                # Keep finished pods around for ten seconds. If job is not
                # deleted its PVC claim persists and it causes PV to stay
                # around.
                # This can be changed by running a cron job that periodically
                # checks for PVC that can be deleted.
                "ttlSecondsAfterFinished": 300,
                "template": {
                    "metadata": {
                        "name": sanitize_kubernetes_label(container_name),
                        "labels": {
                            "app": "resolwe",
                            "data_id": str(data.pk),
                            "process": sanitize_kubernetes_label(data.process.slug),
                            "job_type": sanitize_kubernetes_label(job_type),
                        },
                        "annotations": annotations,
                    },
                    "spec": {
                        "affinity": {},
                        "hostNetwork": use_host_network,
                        "volumes": self._volumes(
                            data.id, location_subpath, core_api, tools_configmaps
                        ),
                        "initContainers": [
                            {
                                "name": sanitize_kubernetes_label(
                                    f"{container_name}-init"
                                ),
                                "image": communicator_image,
                                "imagePullPolicy": "Always",
                                "workingDir": "/",
                                "command": ["/usr/local/bin/python3"],
                                "args": ["-m", "executors.init_container"],
                                "securityContext": {"privileged": True},
                                "volumeMounts": self._init_container_mountpoints(),
                                "env": container_environment,
                            },
                        ],
                        "containers": [
                            {
                                "name": sanitize_kubernetes_label(container_name),
                                "image": processing_container_image,
                                "resources": {"limits": limits, "requests": requests},
                                "securityContext": security_context,
                                "env": container_environment,
                                "workingDir": os.fspath(constants.PROCESSING_VOLUME),
                                "imagePullPolicy": "Always",
                                "command": ["/usr/bin/python3"],
                                "args": ["/processing.py"],
                                "volumeMounts": self._processing_mountpoints(
                                    location_subpath, tools_configmaps
                                ),
                            },
                            {
                                "name": sanitize_kubernetes_label(
                                    f"{container_name}-communicator"
                                ),
                                "image": communicator_image,
                                "imagePullPolicy": "Always",
                                "resources": {
                                    "limits": communicator_limits,
                                    "requests": communicator_requests,
                                },
                                "securityContext": security_context,
                                "env": container_environment,
                                "command": ["/usr/local/bin/python3"],
                                "args": ["/startup.py"],
                                "volumeMounts": self._communicator_mountpoints(
                                    location_subpath
                                ),
                            },
                        ],
                        "restartPolicy": "Never",
                    },
                },
                "backoffLimit": 0,
            },
        }
        self.optimize_job_scheduling(data, job_description)
        start_time = time.time()

        processing_name = constants.PROCESSING_VOLUME_NAME
        input_name = constants.INPUTS_VOLUME_NAME
        if self._should_create_pvc(storage_settings.FLOW_VOLUMES[processing_name]):
            claim_name = unique_volume_name(
                storage_settings.FLOW_VOLUMES[processing_name]["config"]["name"],
                data.id,
            )
            claim_size = limits.pop("storage", 200) * (2**30)  # Default 200 gibibytes
            core_api.create_namespaced_persistent_volume_claim(
                body=self._persistent_volume_claim(
                    claim_name,
                    claim_size,
                    storage_settings.FLOW_VOLUMES[processing_name]["config"],
                ),
                namespace=self.kubernetes_namespace,
                _request_timeout=KUBERNETES_TIMEOUT,
            )
        if input_name in storage_settings.FLOW_VOLUMES:
            if self._should_create_pvc(storage_settings.FLOW_VOLUMES[input_name]):
                claim_size = self._data_inputs_size(data)
                claim_name = unique_volume_name(
                    storage_settings.FLOW_VOLUMES[input_name]["config"]["name"],
                    data.id,
                )
                core_api.create_namespaced_persistent_volume_claim(
                    body=self._persistent_volume_claim(
                        claim_name,
                        claim_size,
                        storage_settings.FLOW_VOLUMES[input_name]["config"],
                    ),
                    namespace=self.kubernetes_namespace,
                    _request_timeout=KUBERNETES_TIMEOUT,
                )

        logger.debug(f"Creating namespaced job: {job_description}")
        batch_api.create_namespaced_job(
            body=job_description,
            namespace=self.kubernetes_namespace,
            _request_timeout=KUBERNETES_TIMEOUT,
        )
        end_time = time.time()
        logger.info(
            "It took {:.2f}s to send config to kubernetes".format(end_time - start_time)
        )

    def _generate_container_name(self, prefix: str, data_id: int):
        """Generate unique container name.

        Name of the kubernetes container should contain only lower case
        alpfanumeric characters and dashes. Underscores are not allowed.
        """
        return "{}-{}".format(prefix, data_id)

    def submit(self, data: Data, argv):
        """Run process.

        For details, see
        :meth:`~resolwe.flow.managers.workload_connectors.base.BaseConnector.submit`.
        """
        (host, port, protocol) = argv[-1].rsplit(" ", maxsplit=3)[-3:]
        self._initialize_variables()
        self.start(data, (host, port, protocol))

        logger.debug(
            __(
                "Connector '{}' running for Data with id {} ({}).",
                self.__class__.__module__,
                data.id,
                repr(argv),
            )
        )

    def cleanup(self, data_id: int):
        """Remove the persistent volume claims created by the executor."""
        try:
            self._load_kubernetes_config()
        except kubernetes.config.config_exception.ConfigException as exception:
            logger.exception("Could not load the kubernetes configuration.")
            raise exception

        core_api = kubernetes.client.CoreV1Api()
        claim_names = [
            unique_volume_name(type_, data_id)
            for type_ in [
                constants.PROCESSING_VOLUME_NAME,
                constants.INPUTS_VOLUME_NAME,
            ]
            if type_ in storage_settings.FLOW_VOLUMES
            and self._should_create_pvc(storage_settings.FLOW_VOLUMES[type_])
        ]
        for persistent_claim_name in claim_names:
            logger.debug("Kubernetes: removing claim %s.", persistent_claim_name)
            with suppress(kubernetes.client.rest.ApiException):
                core_api.delete_namespaced_persistent_volume_claim(
                    name=persistent_claim_name,
                    namespace=self.kubernetes_namespace,
                    _request_timeout=KUBERNETES_TIMEOUT,
                )



"""Process runtime."""
import logging
import sys
from pathlib import Path
from typing import Dict, List, Union

from .communicator import communicator
from .descriptor import ProcessDescriptor
from .fields import Field, RelationDescriptor
from .models import Data, JSONDescriptor

# Inputs class name.
PROCESS_INPUTS_NAME = "Input"
# Outputs class name.
PROCESS_OUTPUTS_NAME = "Output"
logger = logging.getLogger(__name__)


class FlushOnWrite:
    """Flush stream after every write.

    This is a middle ground between buffered stdout/stderr Python uses when it
    is connected to a pipe and unbuffered stdout/stderrr. The first one couses
    that stdout in not updated in real time (and lost on crashes) and the
    second solution causes stdout update for every character.
    """

    def __init__(self, stream):
        """Store the stream for later use."""
        self.stream = stream

    def write(self, data):
        """Write data to the stream and flush it."""
        self.stream.write(data)
        self.stream.flush()

    def writelines(self, datas):
        """Write lines to the stream and flush it."""
        self.stream.writelines(datas)
        self.stream.flush()

    def __getattr__(self, attr):
        """Redirect other attribute lookups to the underlying stream."""
        return getattr(self.stream, attr)


sys.stdout = FlushOnWrite(sys.stdout)  # type: ignore
sys.stderr = FlushOnWrite(sys.stderr)  # type: ignore


class ProcessMeta(type):
    """Metaclass for process.

    It is used to construct ProcessDescriptor for each class that inherites
    from Process. That information is used to generate the process schema.
    """

    def __new__(mcs, name, bases, namespace, **kwargs):
        """Create new process class."""
        # If this an abstract process (ie the parent class for real processes)
        # construct and return its instance immediatelly.
        if "_abstract" in namespace:
            return type.__new__(mcs, name, bases, namespace)

        # Generate a runtime version of the process descriptor.
        meta = ProcessDescriptor()
        for meta_name in dir(meta.metadata):
            if meta_name.startswith("_"):
                continue

            setattr(meta.metadata, meta_name, namespace.pop(meta_name, None))

        for nsp, fields in (
            (PROCESS_INPUTS_NAME, meta.inputs),
            (PROCESS_OUTPUTS_NAME, meta.outputs),
        ):
            inputs = namespace.pop(nsp, {})
            for field_name in dir(inputs):
                if field_name.startswith("_"):
                    continue

                field = getattr(inputs, field_name)
                if not isinstance(field, Field):
                    continue
                field.contribute_to_class(meta, fields, field_name)

        meta.validate()

        result = type.__new__(mcs, name, bases, namespace)
        result._meta = meta
        return result


class Process(metaclass=ProcessMeta):
    """Resolwe process.

    This is the base class from which all other python processes must inherit.
    When one wants to create an abstract python process (a process which is
    not automatically registered in the Process database table) the class
    attribute `_abstract` must be set to True.
    """

    # Mark it an abstract process: such processes are not registered in the
    # database and are parents for actual processes.
    # Make sure to add this class property to all process parent classes.
    _abstract = True

    def __init__(self, data: Data):
        """Construct a new process instance."""
        self.logger = logging.getLogger(self.__class__.__name__)
        self.data = data
        self.process = self.data.process

    def run(self, inputs: JSONDescriptor, outputs: JSONDescriptor):
        """Process entry point."""
        raise NotImplementedError

    def export_files(self, exports: Union[str, List, Dict]):
        """Export files."""

        def file_list(value):
            values = []
            if isinstance(value, str) and Path(value).is_file():
                # TODO: Use the protocol to export files and get the
                # process schema to check field type.
                values = [value]
            elif isinstance(value, dict):
                values = []
                for item in value.values():
                    values += file_list(item)
            elif isinstance(value, list):
                values = []
                for item in value:
                    values += file_list(item)
            return values

        files = file_list(exports)
        communicator.export_files(files)

    def run_process(self, slug: str, inputs: Dict):
        """Run a new process from a running process."""
        self.export_files(inputs)
        communicator.run({"process": slug, "input": inputs})

    def progress(self, progress: float):
        """Report process progress.

        :raises AssertionError: when float is not between 0 and 1.
        """
        assert 0 <= progress <= 1, "Progress must be a float between 0 and 1."
        communicator.progress(round(progress * 100))

    def _process_log(self, log: Dict[str, List[str]]):
        """Send process log.

        The log may contain multiple info, warning and error messages.

        :param log: dictionary with keys 'info', 'warning' and 'error'. The
            corresponding values are lists of strings. Some keys may be
            missing.
        """
        communicator.process_log(log)

    def info(self, *args):
        """Log informational message."""
        to_send = " ".join([str(x) for x in args])
        self._process_log({"info": [to_send]})

    def warning(self, *args):
        """Log warning message."""
        to_send = " ".join([str(x) for x in args])
        self._process_log({"warning": [to_send]})

    def error(self, *args):
        """Log error message."""
        to_send = " ".join([str(x) for x in args])
        self._process_log({"error": [to_send]})

    def get_data_id_by_slug(self, slug) -> int:
        """Find data object ID for given slug."""
        return Data.from_slug(slug).id

    def update_entity_descriptor(self, annotations: dict):
        """Update entity descriptor from dictionary of annotations."""
        communicator.annotate(annotations)

    @property
    def requirements(self):
        """Process requirements."""

        class dotdict(dict):
            """Dot notation access to dictionary attributes."""

            def __getattr__(self, attr):
                value = self.get(attr)
                return dotdict(value) if isinstance(value, dict) else value

        return dotdict(communicator.get_self_requirements(self.data.process.id))

    @property
    def name(self):
        """Get Data name."""
        return self.data.name

    @property
    def relations(self):
        """Process relations."""
        relations = communicator.get_relations(self.data.id)
        return [RelationDescriptor.from_dict(data) for data in relations]

    def start(self):
        """Start the process.

        :param inputs: An instance of `Inputs` describing the process inputs
        :return: An instance of `Outputs` describing the process outputs
        """
        self.logger.info("Process is starting")
        try:
            self.run(self.data.input, self.data.output)
            return self.data.output.freeze()
        except Exception as error:
            self.logger.exception("Exception while running process")
            self.error(str(error))
            raise
        except:  # noqa
            self.logger.exception("Exception while running process")
            self.error("Exception while running process")
            raise
        finally:
            self.logger.info("Process has finished")



# pylint: disable=missing-docstring
import os
import sys
import unittest

from django.test import LiveServerTestCase, override_settings

import resolwe.permissions.models
from resolwe.flow.models import (
    Collection,
    Data,
    DescriptorSchema,
    Entity,
    Process,
    Relation,
    RelationPartition,
    RelationType,
    Storage,
)
from resolwe.flow.models.annotations import (
    AnnotationField,
    AnnotationGroup,
    AnnotationType,
    AnnotationValue,
)
from resolwe.permissions.models import Permission, get_anonymous_user
from resolwe.test import (
    ProcessTestCase,
    tag_process,
    with_docker_executor,
    with_resolwe_host,
)

PROCESSES_DIR = os.path.join(os.path.dirname(__file__), "processes")
WORKFLOWS_DIR = os.path.join(os.path.dirname(__file__), "workflows")
DESCRIPTORS_DIR = os.path.join(os.path.dirname(__file__), "descriptors")
FILES_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), "files")


class PythonProcessTest(ProcessTestCase):
    def setUp(self):
        super().setUp()
        self._register_schemas(
            processes_paths=[PROCESSES_DIR, WORKFLOWS_DIR],
            descriptors_paths=[DESCRIPTORS_DIR],
        )
        self.files_path = FILES_PATH
        # Force reading anonymous user from the database for every test.
        resolwe.permissions.models.ANONYMOUS_USER = None

    @with_docker_executor
    def test_registration(self):
        process = Process.objects.get(slug="test-python-process")

        self.assertEqual(process.slug, "test-python-process")
        self.assertEqual(process.name, "Test Python Process")
        self.assertEqual(process.version, "0.1.2")
        self.assertEqual(process.type, "data:python:")
        self.assertEqual(process.category, "analyses:")
        self.assertEqual(process.scheduling_class, Process.SCHEDULING_CLASS_BATCH)
        self.assertEqual(process.persistence, Process.PERSISTENCE_CACHED)
        self.assertEqual(process.description, "This is a process description.")
        self.assertEqual(process.data_name, "Foo: {{input_data | name}}")
        self.assertEqual(process.entity_type, "sample")
        self.assertEqual(process.entity_descriptor_schema, "sample")
        self.assertEqual(process.entity_input, "input_data")
        self.assertEqual(
            process.requirements,
            {
                "expression-engine": "jinja",
                "executor": {
                    "docker": {
                        "image": "public.ecr.aws/s4q6j6e8/resolwe/base:ubuntu-20.04"
                    }
                },
            },
        )

        for field in process.input_schema:
            if field["name"] == "my_group":
                self.assertEqual(field["group"][0]["name"], "foo")
                self.assertEqual(field["group"][1]["name"], "bar")
                break
        else:
            self.fail("Field my_group not found in test-python-process")

        for field in process.input_schema:
            if field["name"] == "bar":
                self.assertEqual(field["relation"]["type"], "group")
                self.assertEqual(field["relation"]["npartitions"], "none")
                break
        else:
            self.fail("Field bar not found in test-python-process")

        for field in process.input_schema:
            if field["name"] == "baz":
                self.assertEqual(field["relation"]["type"], "group")
                self.assertEqual(field["relation"]["npartitions"], 1)
                break
        else:
            self.fail("Field baz not found in test-python-process")

        for field in process.input_schema:
            if field["name"] == "baz_list":
                self.assertEqual(field["relation"]["type"], "group")
                self.assertEqual(field["relation"]["npartitions"], 1)
                break
        else:
            self.fail("Field baz_list not found in test-python-process")

        for field in process.input_schema:
            if field["name"] == "my_float":
                self.assertEqual(field["range"], [0.0, 1.0])
                break
        else:
            self.fail("Field my_float not found in test-python-process")
        for field in process.input_schema:
            if field["name"] == "integer":
                self.assertEqual(field["range"], [0, 100])
                break
        else:
            self.fail("Field integer not found in test-python-process")

        # Make sure that process with inheritance from `module.Class` is also registered.
        process = Process.objects.get(slug="test-python-process-2")

    @with_docker_executor
    @tag_process("test-python-process-annotate-entity-v2")
    @tag_process("test-python-process-annotate-entity-v2-bulk-update")
    @tag_process("test-python-process-annotate-entity-v2-bulk-set")
    @tag_process("test-python-process-update-entity-annotations-v2")
    def test_annotation_v2(self):
        group = AnnotationGroup.objects.create(name="general", sort_order=1)
        AnnotationField.objects.create(
            name="species", sort_order=1, group=group, type=AnnotationType.STRING.value
        )
        age_field = AnnotationField.objects.create(
            name="age", sort_order=1, group=group, type=AnnotationType.INTEGER.value
        )
        data = self.run_process("test-python-process-annotate-entity-v2")
        self.assertIsNotNone(data.entity)
        self.assertEqual(data.entity.annotations.count(), 2)
        self.assertAnnotation(data.entity, "general.species", "Human")
        self.assertAnnotation(data.entity, "general.age", 42)

        # Try bulk updating.
        data = self.run_process("test-python-process-annotate-entity-v2-bulk-update")
        self.assertIsNotNone(data.entity)
        self.assertEqual(data.entity.annotations.count(), 2)
        self.assertAnnotation(data.entity, "general.species", "Human Bulk")
        self.assertAnnotation(data.entity, "general.age", 42 * 2)

        # Try bulk setting.
        data = self.run_process("test-python-process-annotate-entity-v2-bulk-set")
        self.assertIsNotNone(data.entity)
        self.assertEqual(data.entity.annotations.count(), 1)
        self.assertAnnotation(data.entity, "general.species", "Human Bulk Set")

        # Now try reading and updating existing annotations.
        entity = Entity.objects.create(name="Entity", contributor=self.contributor)
        entity.set_permission(Permission.EDIT, self.contributor)
        AnnotationValue.objects.create(entity=entity, field=age_field, value=42)
        process = Process.objects.get(
            slug="test-python-process-update-entity-annotations-v2"
        )
        process.set_permission(Permission.VIEW, self.contributor)
        data = self.run_process(
            "test-python-process-update-entity-annotations-v2",
            {"entity_id": entity.pk},
            contributor=self.contributor,
        )
        self.assertEqual(data.output["existing_annotations"], "{'general.age': 42}")
        self.assertAnnotation(entity, "general.species", "Human")
        self.assertAnnotation(entity, "general.age", 21)

    @with_docker_executor
    @tag_process("test-python-process", "test-save-file", "entity-process")
    def test_python_process(self):
        with self.preparation_stage():
            input_data = self.run_process(
                "test-save-file", {"input_file": "testfile.txt"}
            )
            input_data.name = "bar"
            input_data.save()

            input_entity = self.run_process("entity-process")

            storage = input_data.storages.create(
                name="storage", contributor=self.user, json={"value": 42}
            )

        data = self.run_process(
            "test-python-process",
            {
                "my_field": "bar",
                "my_list": ["one", "two", "three"],
                "bar": input_data.pk,
                "url": {"url": "https://www.genialis.com"},
                "input_data": input_data.pk,
                "input_entity_data": input_entity.pk,
                "integer": 42,
                "my_float": 0.42,
                "my_json": storage.pk,
                "my_group": {
                    "bar": "my string",
                    "foo": 21,
                },
            },
        )

        self.assertFields(data, "string_output", "OK")
        self.assertFields(data, "list_string_output", ["foo", "bar"])
        self.assertFile(data, "file_output", "testfile.txt")
        self.assertFiles(data, "list_file_output", ["testfile.txt", "testfile2.txt"])

        # Non-deterministic output.
        del data.output["dir_output"]["size"]
        del data.output["dir_output"]["total_size"]

        self.assertEqual(data.output["dir_output"], {"dir": "test/"})
        self.assertEqual(data.output["input_data_name"], "bar")
        self.assertEqual(data.output["input_entity_name"], "Data with entity")
        self.assertEqual(
            data.output["docker_image"],
            "public.ecr.aws/s4q6j6e8/resolwe/base:ubuntu-20.04",
        )
        self.assertEqual(data.name, "Foo: bar")

        entity = Entity.objects.get(data=data)
        self.assertEqual(entity.data.first(), data)

    @with_docker_executor
    @tag_process("test-python-process-group-field")
    def test_python_process_group(self):
        # Run with explicitly given inputs.
        data = self.run_process(
            "test-python-process-group-field",
            {
                "my_group": {
                    "foo": 123,
                    "bar": "foobar",
                },
                "my_group2": {
                    "foo": 124,
                },
                "my_subgroup": {"subgroup": {"foo": 3}},
            },
        )
        self.assertFields(data, "out_foo", 123)
        self.assertFields(data, "out_bar", "foobar")
        self.assertFields(data, "out_foo2", 124)
        self.assertFields(data, "out_subgroup", 3)

        # Run with no inputs - check that default values are used.
        data = self.run_process("test-python-process-group-field")
        self.assertFields(data, "out_foo", 42)
        self.assertFalse(hasattr(data.output, "out_bar"))
        self.assertFalse(hasattr(data.output, "out_bar2"))
        self.assertFields(data, "out_subgroup", 2)

    @with_docker_executor
    @tag_process("test-python-process-json")
    def test_python_process_json(self):
        """Test that data object with json output can be given as input."""
        with self.preparation_stage():
            input_data = self.run_process("test-output-json")

        self.run_process(
            "test-python-process-json",
            {
                "data": input_data.pk,
            },
        )

    @with_docker_executor
    @tag_process("test-non-required-data-inputs")
    def test_non_required_data_input(self):
        """Test workflow with non-required data inputs"""
        with self.preparation_stage():
            input_data = self.run_process("test-output-json")

        self.run_process(
            "test-non-required-data-inputs",
            {
                "data": input_data.pk,
            },
        )

        data = Data.objects.get(process__slug="test-python-process-json")
        self.assertEqual(data.status, "OK")

    @with_docker_executor
    @tag_process("process-with-workflow-input")
    def test_workflow_as_list_input(self):
        """Test workflow with non-required data inputs"""
        with self.preparation_stage():
            workflow = self.run_process("simple-workflow")

        data = self.run_process("process-with-workflow-input", {"data": workflow.pk})

        data.refresh_from_db()
        self.assertEqual(data.status, "OK")

    @with_docker_executor
    @tag_process("test-python-process-error")
    def test_error(self):
        """Test process that raises exception"""
        data = self.run_process(
            "test-python-process-error", assert_status=Data.STATUS_ERROR
        )
        self.assertEqual(data.process_error[0], "Value error in ErrorProcess")

    @with_docker_executor
    @tag_process("test-python-process-file")
    def test_import_file(self):
        """Test import file"""
        inputs = {"src": "testfile.txt"}
        data = self.run_process("test-python-process-file", inputs)
        self.assertEqual(data.output["dst"]["file"], "testfile.txt")
        self.assertEqual(data.output["dst"]["size"], 15)

    @with_docker_executor
    @tag_process("process-with-choices-input")
    def test_process_with_choices(self):
        """Test process that does not have a predefined choice as an input."""
        data = self.run_process("process-with-choices-input", {"string_input": "baz"})
        self.assertFields(data, "string_output", "baz")

    @with_docker_executor
    @tag_process("test-process-relations")
    def test_python_process_relations(self):
        """Test relations in Python process.

        Make two Data (with corresponding entities) in series relation.
        """
        collection = Collection.objects.create(
            name="collection", contributor=self.contributor
        )

        with self.preparation_stage():
            # From collection 1
            start = self.run_process("entity-process")
            end = self.run_process("entity-process")

            start.collection = collection
            end.collection = collection
            start.save()
            end.save()

            # Set relation between the start and end object's entities.
            rel_type_series = RelationType.objects.create(name="series", ordered=True)
            relation = Relation.objects.create(
                contributor=self.contributor,
                collection=collection,
                type=rel_type_series,
                category="time-series",
                unit=Relation.UNIT_HOUR,
            )

            RelationPartition.objects.create(
                relation=relation,
                entity=start.entity,
                label="start",
                position=1,
            )
            RelationPartition.objects.create(
                relation=relation,
                entity=end.entity,
                label="end",
                position=2,
            )

            # Prepare also another data that is not inside entity and is in another collection
            other = self.run_process("test-python-process-2")
            collection_2 = Collection.objects.create(
                name="Collection 2", contributor=self.contributor
            )
            collection_2.data.add(other)

        data = self.run_process(
            "test-process-relations", {"data": [start.pk, end.pk, other.pk]}
        )

        data.refresh_from_db()
        self.assertEqual(data.output["relation_id"], relation.id)
        self.assertEqual(data.output["relation_type"], "series")
        self.assertEqual(data.output["relation_ordered"], "True")
        self.assertEqual(data.output["relation_category"], "time-series")
        self.assertEqual(data.output["relation_unit"], "hr")
        self.assertEqual(data.output["relation_partition_label"], "start")
        self.assertEqual(data.output["relation_partition_position"], 1)

    @with_docker_executor
    @tag_process("data-name-process")
    def test_data_name(self):
        """Test self.name property."""
        with self.preparation_stage():
            data_input = self.run_process("entity-process")

        data = self.run_process("data-name-process", {"data_input": data_input.id})
        self.assertEqual(data.output["name"], "Data with entity")

    @with_docker_executor
    @tag_process("create-collection")
    def test_create_collection(self):
        """Test process that creates object"""
        collection_name = "Python process collection"
        self.assertFalse(Collection.objects.filter(name=collection_name).exists())
        self.run_process("create-collection", {"collection_name": collection_name})
        Collection.objects.get(name=collection_name)

    @with_docker_executor
    @tag_process("storage-objects-test")
    def test_storage_objects(self):
        """Test storage access from python process."""
        data = self.run_process("storage-objects-test")
        self.assertEquals(data.storages.count(), 2)
        storage_string = Storage.objects.get(pk=data.output["output_string"])
        self.assertEqual(storage_string.json, ["valid", "json"])
        storage_file = Storage.objects.get(pk=data.output["output_file"])
        self.assertEqual(storage_file.json, ["valid", "json", "file"])

    @with_docker_executor
    @tag_process("filter-collection")
    def test_filter_collection(self):
        """Test process that filters object"""
        collection_name = "Python process collection"
        number_of_collections = 2
        self.assertFalse(Collection.objects.filter(name=collection_name).exists())
        for _ in range(number_of_collections):
            Collection.objects.create(name=collection_name, contributor=self.user)
        data = self.run_process(
            "filter-collection", {"collection_name": collection_name}
        )
        self.assertEqual(data.output["number_of_collections"], number_of_collections)

    @with_docker_executor
    @tag_process("get-collection")
    def test_get_collection(self):
        """Test process that gets object"""
        collection_name = "Python process collection"
        self.assertFalse(Collection.objects.filter(name=collection_name).exists())
        collection = Collection.objects.create(
            name=collection_name, contributor=self.user
        )
        data = self.run_process("get-collection", {"collection_name": collection_name})
        self.assertEqual(data.output["collection_slug"], collection.slug)

    @with_docker_executor
    @tag_process("parent-process-schema", "test-python-process-requirements")
    def test_get_parent_schema(self):
        """Test access to parent object schema.

        Even when user has no permission to access the process of the parent
        data object.
        """
        anonymous_user = get_anonymous_user()
        main_process = Process.objects.get(slug="parent-process-schema")
        main_process.set_permission(Permission.VIEW, anonymous_user)
        input_process = Process.objects.get(slug="test-python-process-requirements")
        input_process.set_permission(Permission.NONE, anonymous_user)

        inputs = {
            "input1": self.run_process(input_process.slug).id,
            "input2": self.run_process(input_process.slug).id,
        }
        self.run_process(main_process.slug, inputs, contributor=anonymous_user)

    @with_docker_executor
    @tag_process("get-latest-process")
    def test_get_process(self):
        """Test process that gets process by slug.

        The catch here is that multiple objects are returned, as there can be
        multiple processes with the same slug but different versions. In such
        case the object with the process with the latest version must be
        returned.
        """
        process_slug = "multiple-versions"
        self.assertFalse(Process.objects.filter(slug=process_slug).exists())

        process1 = Process.objects.create(
            persistence=Process.PERSISTENCE_TEMP,
            version="1.0.0",
            slug=process_slug,
            contributor=self.admin,
        )
        data = self.run_process("get-latest-process", {"process_slug": process_slug})
        self.assertEqual(data.output["process_pk"], process1.pk)

        process2 = Process.objects.create(
            contributor=self.admin,
            persistence=Process.PERSISTENCE_TEMP,
            version="1.0.1",
            slug=process_slug,
        )
        data = self.run_process("get-latest-process", {"process_slug": process_slug})
        self.assertEqual(data.output["process_pk"], process2.pk)

    @with_docker_executor
    @tag_process("create-data")
    def test_create_data(self):
        """Test process that creates data object."""
        collection_name = "Python process collection"
        data_name = "Data name"
        self.assertFalse(Collection.objects.filter(name=collection_name).exists())
        self.run_process(
            "create-data", {"collection_name": collection_name, "data_name": data_name}
        )
        Collection.objects.get(name=collection_name)
        Data.objects.get(name=data_name)

    @with_docker_executor
    @tag_process("assign-entity-tags")
    def test_assign_tags_entity(self):
        """Assign tags to entity."""
        data = self.run_process(
            "assign-entity-tags",
            {
                "data_name": "data_name",
                "sample_name": "sample_name",
                "tags": ["first", "second"],
            },
        )
        self.assertEqual(data.name, "data_name")
        self.assertEqual(data.entity.name, "sample_name")
        self.assertEqual(data.entity.tags, ["first", "second"])

    @with_docker_executor
    @tag_process("change-entity-name")
    def test_change_entity_name(self):
        """Assign tags to entity."""

        entity = Entity.objects.create(name="Entity", contributor=self.user)
        process = Process.objects.get(slug="change-entity-name")
        process.set_permission(Permission.VIEW, self.user)

        data = self.run_process(
            "change-entity-name",
            {"entity_id": entity.pk, "entity_name": "New entity name"},
            contributor=self.user,
            assert_status=Data.STATUS_ERROR,
        )
        self.assertEqual(
            data.process_error,
            ["No objects match the given criteria or no permission to read object."],
        )
        entity.refresh_from_db()
        self.assertEqual(entity.name, "Entity")

        entity.set_permission(Permission.VIEW, self.user)
        data = self.run_process(
            "change-entity-name",
            {"entity_id": entity.pk, "entity_name": "New entity name"},
            contributor=self.user,
            assert_status=Data.STATUS_ERROR,
        )
        self.assertEqual(len(data.process_error), 1)

        self.assertTrue(
            f"No edit permission for entity with id {entity.pk}."
            in data.process_error[0]
        )
        entity.refresh_from_db()
        self.assertEqual(entity.name, "Entity")

        entity.set_permission(Permission.EDIT, self.user)
        data = self.run_process(
            "change-entity-name",
            {"entity_id": entity.pk, "entity_name": "New entity name"},
            contributor=self.user,
        )
        self.assertEqual(len(data.process_error), 0)
        entity.refresh_from_db()
        self.assertEqual(entity.name, "New entity name")

    @with_docker_executor
    @tag_process("change-entity-descriptor")
    def test_change_descriptor(self):
        """Assign descriptor to entity."""

        descriptor_schema = DescriptorSchema.objects.create(
            name="Descriptor schema",
            contributor=self.contributor,
            schema=[
                {
                    "name": "Description",
                    "type": "basic:string:",
                    "default": "default value",
                }
            ],
        )
        entity = Entity.objects.create(
            name="Entity", contributor=self.user, descriptor_schema=descriptor_schema
        )

        self.run_process(
            "change-entity-descriptor",
            {
                "entity_id": entity.pk,
                "description": "New description",
            },
        )
        entity.refresh_from_db()
        self.assertEqual(
            entity.descriptor,
            {"Description": "New description"},
        )


class PythonProcessRequirementsTest(ProcessTestCase):
    def setUp(self):
        super().setUp()
        self._register_schemas(
            processes_paths=[PROCESSES_DIR, WORKFLOWS_DIR],
            descriptors_paths=[DESCRIPTORS_DIR],
        )
        self.files_path = FILES_PATH

    @with_docker_executor
    @tag_process("test-python-process-requirements")
    def test_defaults(self):
        data = self.run_process("test-python-process-requirements")
        self.assertEqual(data.output["cores"], 2)
        self.assertEqual(data.output["memory"], 4096)
        self.assertEqual(data.output["storage"], 200)

    @with_docker_executor
    @override_settings(FLOW_PROCESS_MAX_CORES=1)
    @tag_process("test-python-process-requirements")
    def test_max_cores(self):
        data = self.run_process("test-python-process-requirements")
        self.assertEqual(data.output["cores"], 1)
        self.assertEqual(data.output["memory"], 4096)
        self.assertEqual(data.output["storage"], 200)

    @with_docker_executor
    @override_settings(
        FLOW_PROCESS_RESOURCE_OVERRIDES={
            "memory": {"test-python-process-requirements": 2048},
            "storage": {"test-python-process-requirements": 300},
        }
    )
    @tag_process("test-python-process-requirements")
    def test_resource_environment_override(self):
        data = self.run_process("test-python-process-requirements")
        self.assertEqual(data.output["cores"], 2)
        self.assertEqual(data.output["memory"], 2048)
        self.assertEqual(data.output["storage"], 300)

    @with_docker_executor
    @override_settings(
        FLOW_PROCESS_RESOURCE_OVERRIDES={
            "memory": {"test-python-process-requirements": 2048},
            "storage": {"test-python-process-requirements": 300},
        }
    )
    @tag_process("test-python-process-requirements")
    def test_resource_data_override(self):
        data = self.run_process(
            "test-python-process-requirements",
            process_resources={"cores": 3, "storage": 500, "memory": 50000},
        )
        self.assertEqual(data.output["cores"], 3)
        self.assertEqual(data.output["memory"], 50000)
        self.assertEqual(data.output["storage"], 500)


class PythonProcessDataBySlugTest(ProcessTestCase, LiveServerTestCase):
    def setUp(self):
        super().setUp()
        self._register_schemas(
            processes_paths=[PROCESSES_DIR], descriptors_paths=[DESCRIPTORS_DIR]
        )
        self.files_path = FILES_PATH

    @unittest.skipUnless(
        sys.platform.startswith("linux"),
        "Accessing live Resolwe host from a Docker container on non-Linux systems is not possible yet.",
    )
    @with_resolwe_host
    @with_docker_executor
    @tag_process("test-python-process-data-id-by-slug", "test-python-process-2")
    def test_process_data_by_slug(self):
        """Test that data object with json output can be given as input."""
        with self.preparation_stage():
            input_data = self.run_process("test-python-process-2")

        input_data = Data.objects.get(id=input_data.id)

        data = self.run_process(
            "test-python-process-data-id-by-slug",
            {
                "slug": input_data.slug,
            },
        )

        self.assertEqual(data.output["data_id"], input_data.pk)



# pylint: disable=missing-docstring
import shutil
import unittest
from datetime import timedelta

from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.test import TransactionTestCase as DjangoTransactionTestCase
from django.utils.timezone import now

from rest_framework import exceptions, status

from resolwe.flow.models import Collection, Data
from resolwe.flow.serializers import ContributorSerializer
from resolwe.flow.views import DataViewSet
from resolwe.observers.models import BackgroundTask
from resolwe.permissions.models import Permission
from resolwe.test import ResolweAPITestCase, TransactionResolweAPITestCase

DATE_FORMAT = r"%Y-%m-%dT%H:%M:%S.%f"

MESSAGES = {
    "NOT_FOUND": "Not found.",
    # 'NO_PERMISSION': 'You do not have permission to perform this action.',
    "ONE_ID_REQUIRED": "Exactly one id required on create.",
}


class DataTestCaseCommonMixin:
    fixtures = [
        "users.yaml",
        "permissions.yaml",
        "collections.yaml",
        "processes.yaml",
        "data.yaml",
    ]

    def setUp(self):
        self.data1 = Data.objects.get(pk=1)

        self.resource_name = "data"
        self.viewset = DataViewSet

        self.data = {
            "name": "New data",
            "slug": "new_data",
            "collection": {"id": 1},
            "process": {"slug": "test_process"},
        }

        super().setUp()

    def tearDown(self):
        for data in Data.objects.all():
            if data.location:
                shutil.rmtree(data.location.get_path(), ignore_errors=True)

        super().tearDown()


class DataTestCase(DataTestCaseCommonMixin, ResolweAPITestCase):
    def test_get_list(self):
        resp = self._get_list(self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(len(resp.data), 2)

    def test_get_list_public_user(self):
        resp = self._get_list()
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(len(resp.data), 1)

    def test_get_list_admin(self):
        resp = self._get_list(self.admin)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(len(resp.data), 2)

    @unittest.skipIf(
        True,
        "since PR308: this test uses transactions, incompatible with the separated manager",
    )
    def test_post(self):
        # logged-in user w/ perms
        collection_n = Data.objects.count()

        resp = self._post(self.data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_201_CREATED)
        self.assertEqual(Data.objects.count(), collection_n + 1)

        d = Data.objects.get(pk=resp.data["id"])
        self.assertTrue(now() - d.modified < timedelta(seconds=1))
        self.assertTrue(now() - d.created < timedelta(seconds=1))
        self.assertEqual(d.status, "OK")

        self.assertTrue(now() - d.started < timedelta(seconds=1))
        self.assertTrue(now() - d.finished < timedelta(seconds=1))
        self.assertEqual(d.contributor_id, self.user1.pk)

    def test_post_invalid_fields(self):
        data_n = Data.objects.count()

        self.data["collection"] = {"id": 42}
        resp = self._post(self.data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            resp.data["collection"][0],
            "Invalid collection value: {'id': 42} - object does not exist.",
        )

        self.data["collection"] = {"id": 1}
        self.data["process"] = {"id": 42}
        resp = self._post(self.data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            str(resp.data["process"][0]),
            "Invalid process value: {'id': 42} - object does not exist.",
        )

        self.assertEqual(Data.objects.count(), data_n)

    def test_post_no_perms(self):
        collection = Collection.objects.get(pk=1)
        collection.set_permission(Permission.VIEW, self.user2)

        data_count = Data.objects.count()
        resp = self._post(self.data, self.user2)
        self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(Data.objects.count(), data_count)

    def test_post_public_user(self):
        data_count = Data.objects.count()
        resp = self._post(self.data)
        # User has no permission to add Data object to the collection.
        self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(Data.objects.count(), data_count)

    def test_post_protected_fields(self):
        date_now = now()
        self.data["created"] = date_now - timedelta(days=360)
        self.data["modified"] = date_now - timedelta(days=180)
        self.data["started"] = date_now - timedelta(days=180)
        self.data["finished"] = date_now - timedelta(days=90)
        self.data["checksum"] = "fake"
        self.data["status"] = "DE"
        self.data["process_progress"] = 2
        self.data["process_rc"] = 18
        self.data["process_info"] = "Spam"
        self.data["process_warning"] = "More spam"
        self.data["process_error"] = "Even more spam"
        self.data["contributor_id"] = 2

        resp = self._post(self.data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_201_CREATED)

        self.assertEqual(resp.data["started"], None)
        self.assertEqual(resp.data["finished"], None)
        self.assertEqual(
            resp.data["checksum"],
            "05bc76611c382a88817389019679f35cdb32ac65fe6662210805b588c30f71e6",
        )
        self.assertEqual(resp.data["status"], "RE")
        self.assertEqual(resp.data["process_progress"], 0)
        self.assertEqual(resp.data["process_rc"], None)
        self.assertEqual(resp.data["process_info"], [])
        self.assertEqual(resp.data["process_warning"], [])
        self.assertEqual(resp.data["process_error"], [])
        self.assertEqual(
            resp.data["contributor"],
            {
                "id": self.user1.pk,
                "username": self.user1.username,
                "first_name": self.user1.first_name,
                "last_name": self.user1.last_name,
            },
        )

    def test_post_contributor_numeric(self):
        response = ContributorSerializer(
            ContributorSerializer().to_internal_value(self.user1.pk)
        ).data

        self.assertEqual(
            response,
            {
                "id": self.user1.pk,
                "username": self.user1.username,
                "first_name": self.user1.first_name,
                "last_name": self.user1.last_name,
            },
        )

    def test_post_contributor_dict(self):
        response = ContributorSerializer(
            ContributorSerializer().to_internal_value({"id": self.user1.pk})
        ).data

        self.assertEqual(
            response,
            {
                "id": self.user1.pk,
                "username": self.user1.username,
                "first_name": self.user1.first_name,
                "last_name": self.user1.last_name,
            },
        )

    def test_post_contributor_dict_extra_data(self):
        response = ContributorSerializer(
            ContributorSerializer().to_internal_value(
                {"id": self.user1.pk, "username": "ignored", "first_name": "ignored"}
            )
        ).data

        self.assertEqual(
            response,
            {
                "id": self.user1.pk,
                "username": self.user1.username,
                "first_name": self.user1.first_name,
                "last_name": self.user1.last_name,
            },
        )

    def test_post_contributor_dict_invalid(self):
        with self.assertRaises(exceptions.ValidationError):
            ContributorSerializer().to_internal_value(
                {
                    "invalid-dictionary": True,
                }
            )

    def test_current_user_permissions(self):
        # Remove all permissions.
        self.data1.permission_group.permissions.all().delete()

        # Set permissions.
        self.data1.set_permission(Permission.VIEW, AnonymousUser())
        self.data1.set_permission(Permission.EDIT, self.user1)
        self.data1.set_permission(Permission.SHARE, self.user2)
        self.data1.set_permission(Permission.EDIT, self.admin)

        # Anonymous request, public view.
        expected_permissions = [{"type": "public", "permissions": ["view"]}]
        resp = self._get_detail(self.data1.pk)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertCountEqual(
            resp.data["current_user_permissions"], expected_permissions
        )

        # Anonymous request, no permissions.
        self.data1.set_permission(Permission.NONE, AnonymousUser())
        expected_permissions = [{}]
        resp = self._get_detail(self.data1.pk)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)
        self.data1.set_permission(Permission.VIEW, AnonymousUser())

        # User1 request.
        expected_permissions = [
            {"type": "public", "permissions": ["view"]},
            {
                "type": "user",
                "id": self.user1.pk,
                "name": self.user1.get_full_name() or self.user1.username,
                "username": self.user1.username,
                "permissions": ["view", "edit"],
            },
        ]
        resp = self._get_detail(self.data1.pk, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertCountEqual(
            resp.data["current_user_permissions"], expected_permissions
        )

        # User1 request, no permissions.
        self.data1.set_permission(Permission.NONE, self.user1)
        expected_permissions = [{"type": "public", "permissions": ["view"]}]
        resp = self._get_detail(self.data1.pk, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertCountEqual(
            resp.data["current_user_permissions"], expected_permissions
        )

        # User2 request.
        expected_permissions = [
            {"type": "public", "permissions": ["view"]},
            {
                "type": "user",
                "id": self.user2.pk,
                "name": self.user2.get_full_name() or self.user2.username,
                "username": self.user2.username,
                "permissions": ["view", "edit", "share"],
            },
        ]
        resp = self._get_detail(self.data1.pk, self.user2)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertCountEqual(
            resp.data["current_user_permissions"], expected_permissions
        )

        # Admin request: since admin is superuser all permissions are returned.
        expected_permissions = [
            {"type": "public", "permissions": ["view"]},
            {
                "type": "user",
                "id": self.admin.pk,
                "name": self.admin.get_full_name() or self.admin.username,
                "username": self.admin.username,
                "permissions": ["view", "edit", "share", "owner"],
            },
        ]
        resp = self._get_detail(self.data1.pk, self.admin)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertCountEqual(
            resp.data["current_user_permissions"], expected_permissions
        )

        # Admin request: since admin is superuser all permissions are returned
        # even if there are no permissions in the database.
        self.data1.permission_group.permissions.all().delete()
        expected_permissions = [
            {
                "type": "user",
                "id": self.admin.pk,
                "name": self.admin.get_full_name() or self.admin.username,
                "username": self.admin.username,
                "permissions": ["view", "edit", "share", "owner"],
            },
        ]
        resp = self._get_detail(self.data1.pk, self.admin)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertCountEqual(
            resp.data["current_user_permissions"], expected_permissions
        )

        # Check listing.
        self.data1.permission_group.permissions.all().delete()
        self.data1.set_permission(Permission.VIEW, self.user1)
        data2 = Data.objects.get(pk=2)
        data2.permission_group.permissions.all().delete()
        data2.set_permission(Permission.EDIT, self.user1)
        data2.set_permission(Permission.VIEW, AnonymousUser())
        expected_permissions = {
            self.data1.pk: [
                {
                    "type": "user",
                    "id": self.user1.pk,
                    "name": self.user1.get_full_name() or self.user1.username,
                    "username": self.user1.username,
                    "permissions": ["view"],
                }
            ],
            data2.pk: [
                {
                    "type": "user",
                    "id": self.user1.pk,
                    "name": self.user1.get_full_name() or self.user1.username,
                    "username": self.user1.username,
                    "permissions": ["view", "edit"],
                },
                {"type": "public", "permissions": ["view"]},
            ],
        }

        resp = self._get_list(self.user1)
        self.assertEqual(len(resp.data), 2)
        for i in range(2):
            self.assertCountEqual(
                resp.data[i]["current_user_permissions"],
                expected_permissions[resp.data[i]["id"]],
            )

        # Admin should have all permissions.
        resp = self._get_list(self.admin)
        expected_permissions = {
            self.data1.pk: [
                {
                    "type": "user",
                    "id": self.admin.pk,
                    "name": self.admin.get_full_name() or self.admin.username,
                    "username": self.admin.username,
                    "permissions": ["view", "edit", "share", "owner"],
                }
            ],
            data2.pk: [
                {
                    "type": "user",
                    "id": self.admin.pk,
                    "name": self.admin.get_full_name() or self.admin.username,
                    "username": self.admin.username,
                    "permissions": ["view", "edit", "share", "owner"],
                },
                {"type": "public", "permissions": ["view"]},
            ],
        }
        self.assertEqual(len(resp.data), 2)
        for i in range(2):
            self.assertCountEqual(
                resp.data[i]["current_user_permissions"],
                expected_permissions[resp.data[i]["id"]],
            )

    def test_get_detail(self):
        # public user w/ perms
        resp = self._get_detail(1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(resp.data["id"], 1)

        # user w/ permissions
        resp = self._get_detail(1, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertKeys(
            resp.data,
            [
                "slug",
                "name",
                "created",
                "modified",
                "contributor",
                "started",
                "finished",
                "checksum",
                "status",
                "process",
                "process_progress",
                "process_rc",
                "process_info",
                "process_warning",
                "process_error",
                "input",
                "output",
                "descriptor_schema",
                "descriptor",
                "id",
                "size",
                "scheduled",
                "current_user_permissions",
                "descriptor_dirty",
                "tags",
                "process_memory",
                "process_cores",
                "collection",
                "entity",
                "duplicated",
                "process_resources",
            ],
        )

        # user w/ public permissions
        resp = self._get_detail(1, self.user2)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertKeys(
            resp.data,
            [
                "slug",
                "name",
                "created",
                "modified",
                "contributor",
                "started",
                "finished",
                "checksum",
                "status",
                "process",
                "process_progress",
                "process_rc",
                "process_info",
                "process_warning",
                "process_error",
                "input",
                "output",
                "descriptor_schema",
                "descriptor",
                "id",
                "size",
                "scheduled",
                "current_user_permissions",
                "descriptor_dirty",
                "tags",
                "process_memory",
                "process_cores",
                "collection",
                "entity",
                "duplicated",
                "process_resources",
            ],
        )

    def test_get_detail_no_perms(self):
        # public user w/o permissions
        resp = self._get_detail(2)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(resp.data["detail"], MESSAGES["NOT_FOUND"])

        # user w/o permissions
        resp = self._get_detail(2, self.user2)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(resp.data["detail"], MESSAGES["NOT_FOUND"])

    def test_patch(self):
        data = {"name": "New data"}
        resp = self._patch(1, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.name, "New data")

    def test_patch_no_perms(self):
        data = {"name": "New data"}
        resp = self._patch(2, data, self.user2)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)
        d = Data.objects.get(pk=2)
        self.assertEqual(d.name, "Test data 2")

    def test_patch_public_user(self):
        data = {"name": "New data"}
        resp = self._patch(2, data)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)
        d = Data.objects.get(pk=2)
        self.assertEqual(d.name, "Test data 2")

    def test_patch_protected(self):
        date_now = now()

        # `created`
        resp = self._patch(1, {"created": date_now}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.created.isoformat(), self.data1.created.isoformat())
        self.assertEqual(d.modified.isoformat(), self.data1.modified.isoformat())

        # `modified`
        resp = self._patch(1, {"modified": date_now - timedelta(days=180)}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.modified.isoformat(), self.data1.modified.isoformat())

        # `started`
        resp = self._patch(1, {"started": date_now}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.started.isoformat(), self.data1.started.isoformat())
        self.assertEqual(d.modified.isoformat(), self.data1.modified.isoformat())

        # `finished`
        resp = self._patch(1, {"finished": date_now}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.finished.isoformat(), self.data1.finished.isoformat())
        self.assertEqual(d.modified.isoformat(), self.data1.modified.isoformat())

        # `checksum`
        resp = self._patch(1, {"checksum": "fake"}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(
            d.checksum,
            "05bc76611c382a88817389019679f35cdb32ac65fe6662210805b588c30f71e6",
        )
        self.assertEqual(d.modified.isoformat(), self.data1.modified.isoformat())

        # `status`
        resp = self._patch(1, {"status": "DE"}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.status, "OK")
        self.assertEqual(d.modified.isoformat(), self.data1.modified.isoformat())

        # `process_progress`
        resp = self._patch(1, {"process_progress": 2}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.process_progress, 0)
        self.assertEqual(d.modified.isoformat(), self.data1.modified.isoformat())

        # `process_rc`
        resp = self._patch(1, {"process_rc": 18}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.process_rc, None)
        self.assertEqual(d.modified.isoformat(), self.data1.modified.isoformat())

        # `process_info`
        resp = self._patch(1, {"process_info": "Spam"}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.process_info, [])
        self.assertEqual(d.modified.isoformat(), self.data1.modified.isoformat())

        # `process_warning`
        resp = self._patch(1, {"process_warning": "More spam"}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.process_warning, [])
        self.assertEqual(d.modified.isoformat(), self.data1.modified.isoformat())

        # `process_error`
        resp = self._patch(1, {"process_error": "Even more spam"}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.process_error, [])
        self.assertEqual(d.modified.isoformat(), self.data1.modified.isoformat())

        # `contributor`
        resp = self._patch(1, {"contributor": 2}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.contributor_id, 1)
        self.assertEqual(d.modified.isoformat(), self.data1.modified.isoformat())

        # `process`
        resp = self._patch(1, {"process": 2}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.process_id, 1)
        self.assertEqual(d.modified.isoformat(), self.data1.modified.isoformat())


class DataTestCaseDelete(
    DataTestCaseCommonMixin, TransactionResolweAPITestCase, DjangoTransactionTestCase
):
    def _pre_setup(self, *args, **kwargs):
        """Delete all previously created users.

        The public user is created in the migrations and recreated during test case,
        but with a different id which breaks the loading of the fixtures.
        """
        get_user_model().objects.all().delete()
        super()._pre_setup(*args, **kwargs)

    def test_delete(self):
        response = self._delete(1, self.user1)
        BackgroundTask.objects.get(pk=response.data["id"]).wait()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        query = Data.objects.filter(pk=1).exists()
        self.assertFalse(query)

    def test_delete_no_perms(self):
        response = self._delete(2, self.user2)
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        query = Data.objects.filter(pk=2).exists()
        self.assertTrue(query)

    def test_delete_public_user(self):
        response = self._delete(2)
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        query = Data.objects.filter(pk=2).exists()
        self.assertTrue(query)



"""Command handlers for python processes."""
import importlib
import logging
import os
from base64 import b64encode
from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Type, TypeVar, Union
from zipfile import ZIP_STORED, ZipFile

from django.apps import apps
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.postgres.fields.jsonb import JSONField as JSONFieldb
from django.db.models import ForeignKey, JSONField, ManyToManyField, Model, Q, Value
from django.db.models.functions import Concat

from resolwe.flow.executors import constants
from resolwe.flow.executors.socket_utils import Message, Response, retry
from resolwe.flow.managers.listener.permission_plugin import permission_manager
from resolwe.flow.models import (
    AnnotationField,
    AnnotationValue,
    Collection,
    Entity,
    Process,
)
from resolwe.flow.models.base import UniqueSlugError
from resolwe.flow.models.utils import serialize_collection_relations
from resolwe.flow.utils import dict_dot
from resolwe.storage.connectors import connectors
from resolwe.storage.models import FileStorage
from resolwe.test.utils import is_testing

from .plugin import ListenerPlugin, listener_plugin_manager

if TYPE_CHECKING:
    from resolwe.flow.managers.listener.listener import Processor

logger = logging.getLogger(__name__)
User = get_user_model()

# How many times to retry creating new object on slug collision error.
OBJECT_CREATE_RETRIES = 10

PluginType = TypeVar("PluginType")


class PythonProcess(ListenerPlugin):
    """Handler methods for Python processes."""

    name = "PythonProcess plugin"
    plugin_manager = listener_plugin_manager

    ALLOWED_MODELS_C = ["Collection", "Entity", "Data"]
    ALLOWED_MODELS_RW = ["Data", "Process", "Entity", "Collection", "Storage"]
    WRITE_FIELDS = {"Data.name", "Data.output", "Data.descriptor"}

    def __init__(self):
        """Initialize plugin."""
        self._permission_manager = permission_manager
        self._hydrate_cache: Dict[int, str] = dict()
        super().__init__()

    def handle_resolve_data_path(
        self, data_id: int, message: Message[int], manager: "Processor"
    ) -> Response[str]:
        """Return the base path that stores given data."""
        data_pk = message.message_data
        if data_pk not in self._hydrate_cache or is_testing():
            mount_point = os.fspath(constants.INPUTS_VOLUME)
            data_connectors = FileStorage.objects.get(data__pk=data_pk).connectors
            for connector in connectors.for_storage("data"):
                if connector in data_connectors and connector.mountable:
                    mount_point = f"/data_{connector.name}"
                    break
            self._hydrate_cache[data_pk] = mount_point
        return message.respond_ok(self._hydrate_cache[data_pk])

    def handle_get_python_program(
        self,
        data_id: int,
        message: Message[Tuple[str, Dict[str, Any]]],
        manager: "Processor",
    ) -> Response[str]:
        """Return the process source code for the given data object."""
        run_dict = (
            Process.objects.filter(data__id=data_id)
            .values_list("run", flat=True)
            .last()
        ) or {}
        return message.respond_ok(run_dict.get("program", ""))

    def handle_create_object(
        self,
        data_id: int,
        message: Message[Tuple[str, str, Dict[str, Any]]],
        manager: "Processor",
    ) -> Response[int]:
        """Create new object and return its id.

        :raises RuntimeError: if user has no permission to create the object.
        """

        @retry(
            max_retries=OBJECT_CREATE_RETRIES,
            retry_exceptions=(UniqueSlugError,),
            min_sleep=1,
            max_sleep=1,
        )
        def create_model(model: Type[Model], model_data: Dict[str, Any]):
            """Create the model.

            Retry up to 10 times on slug colision error.
            """
            return model.objects.create(**model_data)

        app_name, model_name, model_data = message.message_data
        full_model_name = f"{app_name}.{model_name}"
        self._permission_manager.can_create(
            manager.contributor(data_id), full_model_name, model_data, data_id
        )
        model = apps.get_model(app_name, model_name)
        model_data["contributor_id"] = manager.contributor(data_id).id
        return message.respond_ok(create_model(model, model_data).id)

    def handle_filter_objects(
        self,
        data_id: int,
        message: Message[
            Union[
                Tuple[str, str, Dict[str, Any], List[str], List[str]],
                Tuple[str, str, Dict[str, Any], List[str]],
            ],
        ],
        manager: "Processor",
    ) -> Response[List[int]]:
        """Get a list of objects based on criteria.

        List is further filtered based on user permissions and sorted according
        to the list of strings the user provides.
        """
        # Sorting was added later. For compatibility reasons handle both
        # message types, remove the one without sorting ASAP.
        sorting: List[str] = []
        if len(message.message_data) == 4:
            app_name, model_name, filters, attributes = message.message_data
        elif len(message.message_data) == 5:
            app_name, model_name, filters, sorting, attributes = message.message_data
        else:
            raise RuntimeError(
                "Message in handle_filter_objects for the object "
                f"{data_id} not in the correct format. Got tuple of"
                f"length {len(message.message_data)}, expected length 4 or 5."
            )

        full_model_name = f"{app_name}.{model_name}"
        model = apps.get_model(app_name, model_name)
        filtered_objects = self._permission_manager.filter_objects(
            manager.contributor(data_id),
            full_model_name,
            model.objects.filter(**filters),
            data_id,
        )
        return message.respond_ok(
            list(filtered_objects.order_by(*sorting).values_list(*attributes))
        )

    def handle_set_entity_annotations(
        self,
        data_id: int,
        message: Message[tuple[int, dict[str, Any], bool]],
        manager: "Processor",
    ):
        """Handle update entity annotation request.

        The first part of the message is the id of the entity and the second one the
        mapping representing annotations. The keys in the are in the format
        f"{group_name}.{field_name}".
        """
        entity_id, annotations, update = message.message_data
        entity = Entity.objects.get(pk=entity_id)
        # Check that the user has the permissions to update the entity.
        self._permission_manager.can_update(
            manager.contributor(data_id), "flow.Entity", entity, {}, data_id
        )
        # Delete all annotations except the ones that are in the request.
        if not update:
            entity.annotations.all().delete()

        annotation_values: list[AnnotationValue] = []
        for field_path, value in annotations.items():
            field = AnnotationField.field_from_path(field_path)
            if field is None:
                raise ValueError(f"Invalid field path: '{field_path}'.")
            value = AnnotationValue(entity_id=entity_id, field=field, value=value)
            # The validation is necessary since values are bulk-created and validation
            # is done inside the save method, which is not called.
            value.validate()
            annotation_values.append(value)
        if annotation_values:
            AnnotationValue.objects.bulk_create(
                annotation_values,
                update_conflicts=True,
                update_fields=["_value"],
                unique_fields=["entity", "field"],
            )
        return message.respond_ok("OK")

    def handle_get_entity_annotations(
        self,
        data_id: int,
        message: Message[tuple[int, Optional[list[str]]]],
        manager: "Processor",
    ) -> Response[dict[str, Any]]:
        """Handle get annotations request.

        The first part of the message is the id of the entity and the second one the
        list representing annotations to retrieve. The annotations are in the format
        f"{group_name}.{field_name}".
        """
        entity_id, annotations = message.message_data
        # Check that the user has the permissions to read the entity.
        entity = self._permission_manager.filter_objects(
            manager.contributor(data_id),
            "flow.Entity",
            Entity.objects.filter(pk=entity_id),
            data_id,
        ).get()
        entity_annotations = entity.annotations
        if annotations is not None:
            if len(annotations) == 0:
                entity_annotations = entity_annotations.none()
            else:
                # Filter only requested annotations. Return only annotations matching group
                # and field name.
                annotation_filter = Q()
                for annotation in annotations:
                    group_name, field_name = annotation.split(".")
                    annotation_filter |= Q(
                        field__group__name=group_name, field__name=field_name
                    )
                entity_annotations.filter(annotation_filter)
        to_return = dict(
            entity_annotations.annotate(
                full_path=Concat("field__group__name", Value("."), "field__name")
            ).values_list("full_path", "_value__value")
        )
        return message.respond_ok(to_return)

    def handle_update_model_fields(
        self,
        data_id: int,
        message: Message[Tuple[str, str, int, Dict[str, Any]]],
        manager: "Processor",
    ) -> Response[str]:
        """Update the value for the given fields.

        The received message format is
        (app_name, model name, model primary key, names -> values).

        Field name can be given in dot notation for JSON fields.

        :raises RuntimeError: if user has no permissions to modify the object.
        """
        app_name, model_name, model_pk, mapping = message.message_data
        full_model_name = f"{app_name}.{model_name}"

        model = apps.get_model(app_name, model_name)
        model_instance = model.objects.get(pk=model_pk)

        self._permission_manager.can_update(
            manager.contributor(data_id),
            full_model_name,
            model_instance,
            mapping,
            data_id,
        )

        # Update all fields except m2m.
        update_fields = []
        for field_name, field_value in mapping.items():
            # Not exactly sure how to handle this. Output is a JSONField and is
            # only updated, other JSON fields should probably be replaced.
            # Compromise: when update is a dict, then only values in dict are
            # updates, else replaced.
            field_meta = model._meta.get_field(field_name)
            if isinstance(field_meta, (JSONField, JSONFieldb)) and isinstance(
                field_value, dict
            ):
                update_fields.append(field_name)
                current_value = getattr(model_instance, field_name)
                for key, value in field_value.items():
                    dict_dot(current_value, key, value)
            elif isinstance(field_meta, ManyToManyField):
                assert isinstance(
                    field_value, list
                ), "Only lists may be assigned to many-to-many relations"
                field = getattr(model_instance, field_name)
                field_value_set = set(field_value)
                current_objects = set(field.all().values_list("pk", flat=True))
                objects_to_add = field_value_set - current_objects
                objects_to_remove = current_objects - field_value_set
                if objects_to_remove:
                    field.remove(*objects_to_remove)
                if objects_to_add:
                    field.add(*objects_to_add)
            # Set ID directly when setting foreign key relations.
            elif isinstance(field_meta, ForeignKey):
                field_name = f"{field_name}_id"
                update_fields.append(field_name)
                setattr(model_instance, field_name, field_value)
            else:
                update_fields.append(field_name)
                setattr(model_instance, field_name, field_value)
        model_instance.save(update_fields=update_fields)
        return message.respond_ok("OK")

    def handle_get_model_fields_details(
        self, data_id: int, message: Message[Tuple[str, str]], manager: "Processor"
    ) -> Response[Dict[str, Tuple[str, bool, Any]]]:
        """Get the field names and types for the given model.

        The response is a dictionary which field maps names to its types.

        :raises RuntimeError: if user has no permission to read the metadata.
        """
        app_name, model_name = message.message_data
        full_model_name = f"{app_name}.{model_name}"

        self._permission_manager.can_read(
            manager.contributor(data_id), full_model_name, data_id
        )
        model = apps.get_model(app_name, model_name)
        response = {}
        for field in model._meta.get_fields():
            related_model = None
            if field.is_relation:
                related_model = field.related_model._meta.label
            response[field.name] = (
                field.get_internal_type(),
                not field.null,
                related_model,
            )
        return message.respond_ok(response)

    def handle_get_model_fields(
        self,
        data_id: int,
        message: Message[Tuple[str, str, int, List[str]]],
        manager: "Processor",
    ) -> Response[Dict[str, Any]]:
        """Return the value of the given model for the given fields.

        The received message format is
        (model_name, primary_key, list_of_fields).

        In case of JSON field the field name can contain underscores to get
        only the part of the JSON we are interested in.
        """
        app_name, model_name, model_pk, field_names = message.message_data
        full_model_name = f"{app_name}.{model_name}"

        model = apps.get_model(app_name, model_name)
        filtered_objects = self._permission_manager.filter_objects(
            manager.contributor(data_id),
            full_model_name,
            model.objects.filter(pk=model_pk),
            data_id,
        )
        values = filtered_objects.values(*field_names)
        # NOTE: non JSON serializable fields are NOT supported. If such field
        # is requested the exception will be handled one level above and
        # response with status error will be returned.
        if values.count() == 1:
            values = values.get()
        else:
            values = dict()
        return message.respond_ok(values)

    def handle_get_relations(
        self, data_id: int, message: Message[int], manager: "Processor"
    ) -> Response[List[dict]]:
        """Get relations for the given collection object."""
        contributor = manager.contributor(data_id)
        collection = (
            Collection.objects.filter(id=message.message_data)
            .filter_for_user(contributor)
            .get()
        )
        return message.respond_ok(serialize_collection_relations(collection))

    def handle_get_process_requirements(
        self, data_id: int, message: Message[int], manager: "Processor"
    ) -> Response[dict]:
        """Return the requirements for the process with the given id."""
        process_id = message.message_data
        filtered_process = Process.objects.filter(pk=process_id).filter_for_user(
            manager.contributor(data_id)
        )[0]

        process_limits = filtered_process.get_resource_limits()
        process_requirements = filtered_process.requirements
        process_requirements["resources"] = process_limits
        return message.respond_ok(process_requirements)

    def handle_get_self_requirements(
        self, data_id: int, message: Message[int], manager: "Processor"
    ) -> Response[dict]:
        """Return the requirements for the process being executed."""
        data = manager.data(data_id)
        limits = data.get_resource_limits()
        process_requirements = data.process.requirements
        process_requirements["resources"] = limits
        return message.respond_ok(process_requirements)

    def handle_get_python_runtime(
        self, data_id: int, message: Message[str], manager: "Processor"
    ) -> Response[str]:
        """Return the Python Process runtime.

        It is gzipped and returned as b64encoded string.
        """

        zipped = BytesIO()
        with ZipFile(file=zipped, mode="w", compression=ZIP_STORED) as zip_handle:
            for runtime_class_name in settings.FLOW_PROCESSES_RUNTIMES:
                module_name, class_name = runtime_class_name.rsplit(".", 1)
                source_path = importlib.util.find_spec(module_name).origin
                assert (
                    source_path is not None
                ), f"Unable to determine the source path of the module {module_name}."
                source_dir = Path(source_path).parent

                base_destination = Path(*module_name.split(".")[:-1])
                for source_entry in source_dir.rglob("*.py"):
                    relative_path = source_entry.relative_to(source_dir)
                    destination = base_destination / relative_path
                    zip_handle.write(source_entry, destination)

                # Create missing __init__.py files with empty content.
                for path in list(Path(*module_name.split(".")).parents)[1:-1]:
                    zip_destination = path / "__init__.py"
                    zip_handle.writestr(os.fspath(zip_destination), "")

        zipped.seek(0)
        return message.respond_ok(b64encode(zipped.read()).decode())

    def handle_get_user_model_label(
        self, data_id: int, message: Message[str], manager: "Processor"
    ) -> Response[str]:
        """Return the label of the custom user model.

        It is used to connect the model in Python process runtime with the
        custom user model.
        """
        return message.respond_ok(User._meta.label)



"""Support for easier dynamic composition of type extensions."""
import inspect
from importlib import import_module

from django.apps import apps


class Composer:
    """Support for easier dynamic composition of type extensions."""

    def __init__(self):
        """Construct new composer instance."""
        self._discovery_done = False
        self._extensions = {}

    def discover_extensions(self):
        """Discover available extensions."""
        if self._discovery_done:
            return

        try:
            previous_state = self._extensions.copy()

            for app_config in apps.get_app_configs():
                indexes_path = "{}.extensions".format(app_config.name)
                try:
                    import_module(indexes_path)
                except ImportError:
                    pass

            self._discovery_done = True
        except Exception:
            # Rollback state to prevent corrupted state on exceptions during import.
            self._extensions = previous_state
            raise

    def _get_class_path(self, klass_or_instance):
        """Return class path for a given class.

        :param klass_or_instance: Class or instance of given class or
            class path as a string
        :return: String containing the class path
        """

        def construct_class_path(klass):
            """Return full path of the class."""
            return "{}.{}".format(klass.__module__, klass.__name__)

        if inspect.isclass(klass_or_instance):
            klass = construct_class_path(klass_or_instance)
        elif not isinstance(klass_or_instance, str):
            klass = construct_class_path(klass_or_instance.__class__)
        else:
            klass = klass_or_instance

        return klass

    def add_extension(self, klass, extension):
        """Register an extension for a class.

        :param klass: Class to register an extension for
        :param extension: Extension (arbitrary type)
        """
        klass = self._get_class_path(klass)

        # TODO: Take order into account.
        self._extensions.setdefault(klass, []).append(extension)

    def get_extensions(self, klass):
        """Return all registered extensions of a class.

        :param klass: Class to get registered extensions for
        :return: All registered extensions for given class
        """
        self.discover_extensions()

        return self._extensions.get(self._get_class_path(klass), [])


composer = Composer()



"""Audit logger."""

import logging
from typing import Optional

from rest_framework.request import Request
from rest_framework.response import Response

from resolwe.auditlog.auditmanager import AccessType, AuditManager, ContentType, Fields

audit_log = logging.getLogger("auditlog")


class AuditLogger:
    """Main audit log class."""

    @property
    def manager(self):
        """Get the global audit manager."""
        return AuditManager.global_instance()

    def info(self, message: str, *args, **kwargs):
        """Log custom message at info level.

        :attr message: the logger message.
        :attr args: arguments to pass to the logger.
        :attr kwargs: keyword arguments to pass to the logger. When 'request'
            or 'response' keys are present it is expected their correspondent
            values are request and response object. The information is
            extracted from them and added to the kwargs.
        """
        extra_data = kwargs.setdefault("extra", dict())
        extra_data.update(
            self._extract_information(kwargs.pop("request"), kwargs.pop("response"))
        )
        audit_log.info(message, *args, **kwargs)

    def _extract_from_request(self, request: Request) -> dict:
        """Extract the information from the request."""
        return {
            "url": request.get_full_path(),
            "server_name": getattr(request.META, "HTTP_HOST", "unknown"),
            "session_id": request.session.session_key
            if hasattr(request, "session")
            else None,
            "request_id": request.META["RESOLWE_AUDIT_MANAGER_REQUEST_ID"],
        }

    def _extract_from_response(self, response: Response) -> dict:
        """Extract the information from the response."""
        return {"status_code": response.status_code}

    def _extract_information(
        self, request: Optional[Request], response: Optional[Response]
    ) -> dict:
        """Extract the information from the request and the response."""
        info = dict()
        if request is not None:
            info.update(self._extract_from_request(request))
        if response is not None:
            info.update(self._extract_from_response(response))
        return info

    def log_object_access(
        self,
        request,
        response,
        content_type: ContentType,
        object_id: int,
        access_type: AccessType,
        fields: Fields,
        level=logging.INFO,
    ):
        """Create an audit log entry.

        :args message: the string to log.
        :args content_type: content_type for audit log. If not given it is
            determined from the queryset.
        :args access_type: C,R,U, or D.
        """
        audit_log_extra_data = {
            "content_type": content_type,
            "accessed_ids": object_id,
            "crud": access_type,
            "fields": ",".join(sorted(fields)),
        }
        audit_log_extra_data.update(self._extract_information(request, response))
        message = (
            "Object accessed: {content_type} {accessed_ids} {fields} {crud}".format(
                **audit_log_extra_data
            )
        )
        audit_log.log(level, message, extra=audit_log_extra_data)


logger = AuditLogger()



"""Mixins for Observable ViewSets."""
from drf_spectacular.utils import extend_schema

from django.contrib.auth.models import User
from django.contrib.contenttypes.models import ContentType

from rest_framework import serializers
from rest_framework.decorators import action
from rest_framework.exceptions import NotFound
from rest_framework.request import Request
from rest_framework.response import Response

from resolwe.permissions.models import get_anonymous_user

from .models import Observer, Subscription
from .protocol import ChangeType


class ObserverSubscribeRequestSerializer(serializers.Serializer):
    """Serializer for the subscribe endpoint."""

    ids = serializers.ListField(
        child=serializers.IntegerField(), required=False, allow_null=True
    )
    session_id = serializers.CharField()


class ObserverSubscriptionSerializer(serializers.Serializer):
    """Serializer for the subscribe response."""

    subscription_id = serializers.UUIDField()


class ObservableMixin:
    """A Mixin to make a model ViewSet observable.

    Add the subscribe and unsubscribe endpoints to the list view.
    """

    def user_has_permission(self, id: int, user: User) -> bool:
        """Verify that an object exists for a given user."""
        return self.get_queryset().filter(pk=id).filter_for_user(user).exists()

    @extend_schema(
        request=ObserverSubscribeRequestSerializer,
        responses={200: ObserverSubscriptionSerializer},
    )
    @action(detail=False, methods=["post"])
    def subscribe(self, request: Request) -> Response:
        """Register an Observer for a resource."""
        deserializer = ObserverSubscribeRequestSerializer(data=request.data)
        deserializer.is_valid(raise_exception=True)
        ids = deserializer.validated_data.get("ids")
        session_id = deserializer.validated_data["session_id"]

        content_type = ContentType.objects.get_for_model(self.get_queryset().model)
        user = request.user if request.user.is_authenticated else get_anonymous_user()
        subscription = Subscription.objects.create(user=user, session_id=session_id)

        if ids is None:
            # Subscribe to the whole table.
            subscription.subscribe(
                content_type, [Observer.ALL_IDS], (ChangeType.CREATE, ChangeType.DELETE)
            )
        else:
            # Verify all ids exists and user has permissions to view them.
            for id in ids:
                if not self.user_has_permission(id, request.user):
                    raise NotFound(f"Item {id} does not exist")

            change_types = (ChangeType.UPDATE, ChangeType.DELETE, ChangeType.CREATE)
            subscription.subscribe(content_type, ids, change_types)

        serializer = ObserverSubscriptionSerializer(
            {"subscription_id": subscription.subscription_id}
        )
        return Response(serializer.data)

    @extend_schema(request=ObserverSubscriptionSerializer, responses={200: None})
    @action(detail=False, methods=["post"])
    def unsubscribe(self, request: Request) -> Response:
        """Unregister a subscription."""
        deserializer = ObserverSubscriptionSerializer(data=request.data)
        deserializer.is_valid(raise_exception=True)
        subscription_id = deserializer.validated_data["subscription_id"]

        user = request.user if request.user.is_authenticated else get_anonymous_user()
        subscription = Subscription.objects.filter(
            subscription_id=subscription_id, user=user
        ).first()

        if subscription is None:
            raise NotFound(
                f"Subscription {subscription_id} for user {user} does not exist."
            )

        subscription.delete()
        return Response()



"""The model Observer model."""
import uuid
from time import sleep, time
from typing import Any, Iterable, List, Optional, Set, Tuple

from asgiref.sync import async_to_sync
from channels.layers import get_channel_layer

from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.db import models, transaction
from django.db.models import Count, Q
from django.db.models.query import QuerySet

from resolwe.permissions.models import Permission, PermissionObject

from .protocol import GROUP_SESSIONS, TYPE_ITEM_UPDATE, ChangeType, ChannelsMessage

# Type alias for observable object.
Observable = PermissionObject


def _default_output_value():
    """Return the default value for output field in the BackgroundTask model.

    Must be named method due to Django bug in migrations.
    """
    return ""


class BackgroundTask(Observable):
    """The observable model representing the background task.

    When a long running task is started, the instance of type BackgroundTask is created
    and frontend can subscribe to receive its notifications.
    """

    #: background task is waiting
    STATUS_WAITING = "WT"
    #: background task is processing
    STATUS_PROCESSING = "PR"
    #: background task has finished successfully
    STATUS_DONE = "OK"
    #: background task has finished with error
    STATUS_ERROR = "ER"

    STATUS_CHOICES = (
        (STATUS_WAITING, "Waiting"),
        (STATUS_PROCESSING, "Processing"),
        (STATUS_DONE, "Done"),
        (STATUS_ERROR, "Error"),
    )

    #: task start date and time
    started = models.DateTimeField(blank=True, null=True, db_index=True)

    #: task finished date and time
    finished = models.DateTimeField(blank=True, null=True, db_index=True)

    #: status of the background task
    status = models.CharField(
        max_length=2, choices=STATUS_CHOICES, default=STATUS_WAITING
    )

    #: task description
    description = models.CharField(max_length=256)

    #: the field containing output from the backgroud task, such as list of ids of
    #: duplicated objects, error details...
    output = models.JSONField(default=_default_output_value)

    def wait(
        self,
        timeout: float = 2,
        final_statuses: list[str] = [STATUS_DONE],
        polling_interval: float = 0.3,
    ):
        """Wait for up to timeout seconds for task to transition into final_statuses.

        The method is meant to be used in tests as it creates lots of database queries.

        :raises RuntimeError: when desired status was not reached within timeout.
        """
        started = time()
        while self.status not in final_statuses and time() - started < timeout:
            sleep(polling_interval)
            self.refresh_from_db()

        if self.status not in final_statuses:
            raise RuntimeError(
                f"Task failed to reach {final_statuses} within {timeout} seconds."
            )

    def result(
        self,
        timeout: float = 2,
        final_statuses: list[str] = [STATUS_DONE],
        polling_interval: float = 0.3,
    ) -> Any:
        """Return the output field of the background task when finished.

        The method waits for up to timeout seconds for task to reach the one of
        final_statuses and returns its result (output field).

        :raises RuntimeError: when desired status was not reached within the timeout.
        """
        self.wait(timeout, final_statuses, polling_interval)
        return self.output


def get_random_uuid() -> str:
    """Generate a random UUID in string format."""
    return uuid.uuid4().hex


class Observer(models.Model):
    """An observer for detecting changes in a model.

    Characterized by a db table or an instance of a model and a change type.
    Several subscriptions can subscribe to the same observer.
    """

    ALL_IDS = 0  # The constant used for catchall object id.
    CHANGE_TYPES = [(change.value, change.name) for change in ChangeType]

    #: table of the observed resource
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    #: primary key of the observed resource (null if watching the whole table)
    object_id = models.PositiveIntegerField(null=False)
    #: the type of change to observe for
    change_type = models.PositiveSmallIntegerField(choices=CHANGE_TYPES)

    class Meta:
        """Add index to session_id field and set defining fields."""

        indexes = [models.Index(fields=["object_id"])]
        unique_together = ("content_type", "object_id", "change_type")

    @classmethod
    def get_interested(
        cls,
        content_type: ContentType,
        object_id: int,
        change_type: Optional[ChangeType] = None,
    ) -> "QuerySet[Observer]":
        """Find all observers watching for changes of a given item/table."""
        query = Q(content_type=content_type)
        if change_type is not None:
            query &= Q(change_type=change_type.value)
        query &= Q(object_id=object_id) | Q(object_id=Observer.ALL_IDS)
        return cls.objects.filter(query)

    @classmethod
    def observe_instance_container(
        cls,
        instance: Observable,
        change_type: ChangeType,
        containers: Optional[Iterable[Observable]] = None,
    ):
        """Handle a notifications to the containers of the given instance.

        When containers are given notify them else infer them from the instance.
        """
        # Test explicitely for None, since containers may be empty.
        if containers is None:
            containers = instance.containers
        for container in containers:
            observers = Observer.get_interested(
                change_type=change_type,
                content_type=ContentType.objects.get_for_model(container),
                object_id=container.pk,
            )
            # Forward the message to the appropriate groups.
            for subscriber in Subscription.objects.filter(observers__in=observers):
                if container.has_permission(Permission.VIEW, subscriber.user):
                    # Register on_commit callbacks to send the signals.
                    Subscription.notify(
                        subscriber.session_id,
                        container,
                        change_type,
                        source=(
                            ContentType.objects.get_for_model(instance).name,
                            instance.pk,
                        ),
                    )

    @classmethod
    def observe_instance_changes(cls, instance: Observable, change_type: ChangeType):
        """Handle a notification about an instance change."""
        content_type = ContentType.objects.get_for_model(instance)
        observers = Observer.get_interested(
            change_type=change_type,
            content_type=content_type,
            object_id=instance.pk,
        )

        # Forward the message to the appropriate groups.
        for subscriber in Subscription.objects.filter(observers__in=observers):
            if instance.has_permission(Permission.VIEW, subscriber.user):
                # Register on_commit callbacks to send the signals.
                Subscription.notify(
                    subscriber.session_id,
                    instance,
                    change_type,
                    source=(content_type.name, instance.pk),
                )

    @classmethod
    def observe_permission_changes(
        cls,
        instance: Any,
        gains: Set[int],
        losses: Set[int],
        observe_containers: bool = True,
    ):
        """Handle a notification about a permission change.

        Given an instance and a set of user_ids who gained/lost permissions for it,
        only relevant observers will be notified of the instance's creation/deletion.

        When observe_containers is set to True (default), also object's containers are
        notified using the same change type.
        """
        for change_type, user_ids in (
            (ChangeType.CREATE, gains),
            (ChangeType.DELETE, losses),
        ):
            # A shortcut if nothing actually changed.
            if len(user_ids) == 0:
                continue

            # Find all sessions who have observers registered on this object.
            content_type = ContentType.objects.get_for_model(instance)
            interested = Observer.get_interested(
                change_type=change_type,
                content_type=content_type,
                object_id=instance.pk,
            )
            # Of all interested users, select only those whose permissions changed.
            session_ids = set(
                Subscription.objects.filter(observers__in=interested)
                .filter(user__in=user_ids)
                .values_list("session_id", flat=True)
                .distinct()
            )

            for session_id in session_ids:
                Subscription.notify(
                    session_id,
                    instance,
                    change_type,
                    source=(content_type.name, instance.pk),
                )

            if observe_containers:
                cls.observe_instance_container(instance, change_type)

    def __str__(self) -> str:
        """Format the object representation."""
        return f"content_type={self.content_type} object_id={self.object_id} change={self.change_type}"


class Subscription(models.Model):
    """Subscription to several observers.

    One subscription corresponds to a single api call to the subscribe endpoint.
    A client may have several subscriptions with different ID's to differentiate
    between them.
    """

    #: observers to whom the subscription is listening
    observers = models.ManyToManyField("Observer", related_name="subscriptions")
    #: subscriber's user reference
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
    )
    #: subscription time
    created = models.DateTimeField(auto_now_add=True)

    #: ID of the websocket session (one session may have multiple observers)
    session_id = models.CharField(max_length=100)
    #: unique ID for the client to remember which subscription a signal belongs to
    subscription_id = models.UUIDField(
        unique=True, default=get_random_uuid, editable=False
    )

    class Meta:
        """Add index to session_id field."""

        indexes = [models.Index(fields=["session_id"])]

    def subscribe(
        self,
        content_type: ContentType,
        object_ids: List[int],
        change_types: List[ChangeType],
    ):
        """Assign self to multiple observers at once."""
        for id in object_ids:
            for change_type in change_types:
                observer, _ = Observer.objects.get_or_create(
                    content_type=content_type,
                    object_id=id,
                    change_type=change_type.value,
                )
                self.observers.add(observer)

    def delete(self):
        """Delete the given subscription.

        Delete all observers with no remaining subscriptions.
        """
        # Find related observers with only one remaining subscription
        # (it must be this one) and delete them first.
        Observer.objects.annotate(subs=Count("subscriptions")).filter(
            subscriptions=self.pk, subs=1
        ).delete()
        super().delete()

    @classmethod
    def notify(
        cls,
        session_id: str,
        instance: Observable,
        change_type: ChangeType,
        source: Optional[Tuple[str, int]],
    ):
        """Register a callback to send a change notification on transaction commit."""
        notification: ChannelsMessage = {
            "type": TYPE_ITEM_UPDATE,
            "content_type_pk": ContentType.objects.get_for_model(instance).pk,
            "change_type_value": change_type.value,
            "object_id": instance.pk,
            "source": source,
        }

        # Define a callback, but copy variable values.
        def trigger(
            channel_layer=get_channel_layer(),
            channel=GROUP_SESSIONS.format(session_id=session_id),
            notification=notification,
        ):
            async_to_sync(channel_layer.group_send)(channel, notification)

        transaction.on_commit(trigger)

    def notify_created(self, content_type: ContentType):
        """Send a create notification.

        Used to send signal when models without permissions are created.
        """
        notification: ChannelsMessage = {
            "type": TYPE_ITEM_UPDATE,
            "content_type_pk": content_type.pk,
            "change_type_value": ChangeType.CREATE.value,
            "object_id": Observer.ALL_IDS,
            "source": None,
        }
        channel_layer = get_channel_layer()
        channel = GROUP_SESSIONS.format(session_id=self.session_id)
        async_to_sync(channel_layer.group_send)(channel, notification)
